
import os
import random as _random
try:
    from typing import Any, Callable
except ImportError:
    pass

from marple.arraymodel import APLArray, S
from marple.backend import (
    _DOWNCAST_CT, is_numeric_array, maybe_downcast, maybe_downcast_scalar,
    maybe_upcast, np, to_list,
)
from marple.cells import clamp_rank, decompose, reassemble, resolve_rank_spec
from marple.errors import APLError, ClassError, DomainError, IndexError_, LengthError, LimitError, RankError, SecurityError, SyntaxError_, ValueError_

# Name classes (following Dyalog ⎕NC convention)
NC_UNKNOWN = 0
NC_ARRAY = 2
NC_FUNCTION = 3
NC_OPERATOR = 4
from marple.functions import (
    absolute_value,
    add,
    ceiling,
    divide,
    equal,
    exponential,
    floor,
    greater_equal,
    greater_than,
    less_equal,
    less_than,
    logarithm,
    logical_and,
    circular,
    logical_not,
    logical_or,
    pi_times,
    maximum,
    minimum,
    multiply,
    natural_log,
    negate,
    not_equal,
    power,
    reciprocal,
    residue,
    subtract,
)
from marple.structural import (
    catenate,
    decode,
    drop,
    encode,
    expand,
    from_array,
    grade_down,
    replicate_first,
    grade_up,
    index_of,
    iota,
    matrix_divide,
    matrix_inverse,
    membership,
    ravel,
    replicate,
    reshape,
    reverse,
    rotate,
    shape,
    take,
    transpose,
)
from marple.namespace import Namespace, load_system_workspace
from marple.parser import (
    Alpha,
    AlphaAlpha,
    AlphaDefault,
    Assignment,
    BoundOperator,
    FmtArgs,
    FunctionRef,
    DerivedFunc,
    Dfn,
    DyadicDfnCall,
    DyadicDopCall,
    DyadicFunc,
    Guard,
    IBeamDerived,
    Index,
    InnerProduct,
    MonadicDfnCall,
    MonadicDopCall,
    MonadicFunc,
    OmegaOmega,
    Nabla,
    Num,
    Omega,
    OuterProduct,
    Program,
    QualifiedVar,
    RankDerived,
    ReduceOp,
    ScanOp,
    Str,
    SysVar,
    Var,
    Vector,
    parse,
)


class _DfnClosure:
    """A dfn paired with its defining environment."""

    def __init__(self, dfn: Dfn, env: dict[str, Any]) -> None:
        self.dfn = dfn
        self.env = env


class _GuardTriggered(Exception):
    """Raised when a guard condition is true to return its value."""

    def __init__(self, value: APLArray) -> None:
        self.value = value


MONADIC_FUNCTIONS: dict[str, object] = {
    "+": lambda omega: omega,
    "-": negate,
    "×": lambda omega: S((-1 if omega.data[0] < 0 else 1 if omega.data[0] > 0 else 0)),
    "÷": reciprocal,
    "⌈": ceiling,
    "⌊": floor,
    "*": exponential,
    "⍟": natural_log,
    "|": absolute_value,
    "~": logical_not,
    "⍴": shape,
    "⍳": iota,
    ",": ravel,
    "⌽": reverse,
    "⍉": transpose,
    "⌹": matrix_inverse,
    "○": pi_times,
}

DYADIC_FUNCTIONS: dict[str, object] = {
    "+": add,
    "-": subtract,
    "×": multiply,
    "÷": divide,
    "⌈": maximum,
    "⌊": minimum,
    "*": power,
    "⍟": logarithm,
    "|": residue,
    "∧": logical_and,
    "∨": logical_or,
    "⍴": reshape,
    ",": catenate,
    "↑": take,
    "↓": drop,
    "⌽": rotate,
    "⊤": encode,
    "⊥": decode,
    "/": replicate,
    "⌿": replicate_first,
    "\\": expand,
    "⌹": matrix_divide,
    "○": circular,
}


def _format_array(omega: APLArray) -> APLArray:
    """Monadic ⍕: format an array as a character vector."""
    from marple.formatting import format_num
    if omega.is_scalar():
        s = format_num(omega.data[0])
    else:
        parts = []
        for val in omega.data:
            parts.append(format_num(val))
        s = " ".join(parts)
    chars = list(s)
    return APLArray([len(chars)], chars)


def _dyadic_format(alpha: APLArray, omega: APLArray) -> APLArray:
    """Dyadic ⍕: format with width (scalar left) or width+precision (2-element left)."""
    if alpha.is_scalar():
        width = int(alpha.data[0])
        precision = None
    else:
        width = int(alpha.data[0])
        precision = int(alpha.data[1]) if len(alpha.data) > 1 else None
    values = omega.data if not omega.is_scalar() else [omega.data[0]]
    result_chars: list[str] = []
    for v in values:
        if precision is not None:
            formatted = f"{float(v):.{precision}f}"
        else:
            formatted = str(v)
        formatted = formatted.rjust(width)
        result_chars.extend(list(formatted))
    return APLArray([len(result_chars)], result_chars)


def _call_dfn(
    closure: _DfnClosure,
    omega: APLArray,
    alpha: APLArray | None = None,
    alpha_alpha: object | None = None,
    omega_omega: object | None = None,
) -> APLArray:
    """Execute a dfn with the given arguments."""
    # Lexical scope: start from the defining environment
    local_env: dict[str, Any] = dict(closure.env)
    local_env["⍵"] = omega
    if alpha is not None:
        local_env["⍺"] = alpha
    if alpha_alpha is not None:
        local_env["⍺⍺"] = alpha_alpha
    if omega_omega is not None:
        local_env["⍵⍵"] = omega_omega
    # Store self-reference for ∇
    local_env["∇"] = closure

    result = S(0)
    try:
        for stmt in closure.dfn.body:
            if isinstance(stmt, AlphaDefault):
                # ⍺←default: set ⍺ only if not already provided
                if "⍺" not in local_env:
                    local_env["⍺"] = _evaluate(stmt.default, local_env)
            elif isinstance(stmt, Guard):
                cond = _evaluate(stmt.condition, local_env)
                if cond.data[0]:
                    raise _GuardTriggered(_evaluate(stmt.body, local_env))
            else:
                result = _evaluate(stmt, local_env)
    except _GuardTriggered as g:
        return g.value
    return result


def _evaluate(node: object, env: dict[str, Any]) -> APLArray:
    if isinstance(node, Num):
        value = node.value
        if isinstance(value, float) and int(env.get("⎕FR", S(645)).data[0]) == 1287:
            from decimal import Decimal
            value = Decimal(str(node.value))
        return S(value)

    if isinstance(node, Str):
        chars = list(node.value)
        return APLArray([len(chars)], chars)

    if isinstance(node, Vector):
        values = [el.value for el in node.elements]
        return APLArray([len(values)], list(values))

    if isinstance(node, Var):
        if node.name not in env:
            raise ValueError_(f"Undefined variable: {node.name}")
        val = env[node.name]
        if isinstance(val, APLArray):
            return val
        if isinstance(val, (_DfnClosure, IBeamDerived)):
            return val  # type: ignore[return-value]
        raise DomainError(f"Unexpected value type for {node.name}: {type(val)}")

    if isinstance(node, QualifiedVar):
        return _resolve_qualified(node.parts, env)

    if isinstance(node, SysVar):
        if node.name == "⎕TS":
            import time
            t = time.localtime()
            return APLArray([7], [t[0], t[1], t[2],
                                  t[3], t[4], t[5], 0])
        if node.name == "⎕VER":
            from marple import __version__
            import sys
            s = "MARPLE v" + __version__ + " on " + sys.platform
            return APLArray([len(s)], list(s))
        if node.name not in env:
            raise ValueError_(f"Undefined system variable: {node.name}")
        return env[node.name]

    if isinstance(node, Index):
        array = _evaluate(node.array, env)
        io = _get_io(env)
        return _bracket_index(array, node.indices, env, io)

    if isinstance(node, Omega):
        if "⍵" not in env:
            raise ValueError_("⍵ used outside of dfn")
        return env["⍵"]

    if isinstance(node, Alpha):
        if "⍺" not in env:
            raise ValueError_("⍺ used outside of dfn")
        return env["⍺"]

    if isinstance(node, AlphaAlpha):
        if "⍺⍺" not in env:
            raise ValueError_("⍺⍺ used outside of dop")
        val = env["⍺⍺"]
        if isinstance(val, APLArray):
            return val
        return val  # type: ignore[return-value]

    if isinstance(node, OmegaOmega):
        if "⍵⍵" not in env:
            raise ValueError_("⍵⍵ used outside of dop")
        val = env["⍵⍵"]
        if isinstance(val, APLArray):
            return val
        return val  # type: ignore[return-value]

    if isinstance(node, Nabla):
        if "∇" not in env:
            raise ValueError_("∇ used outside of dfn")
        return env["∇"]  # type: ignore[return-value]

    if isinstance(node, FunctionRef):
        return node  # type: ignore[return-value]

    if isinstance(node, Dfn):
        return _DfnClosure(node, env)  # type: ignore[return-value]

    if isinstance(node, IBeamDerived):
        return node  # type: ignore[return-value]

    if isinstance(node, MonadicFunc):
        operand = _evaluate(node.operand, env)
        return _dispatch_monadic(node.function, operand, env)

    if isinstance(node, DyadicFunc):
        right = _evaluate(node.right, env)
        left = _evaluate(node.left, env)
        return _dispatch_dyadic(node.function, left, right, env)

    if isinstance(node, MonadicDfnCall):
        if isinstance(node.dfn, SysVar):
            return _call_sys_function_monadic(node.dfn.name, node.operand, env)
        if isinstance(node.dfn, IBeamDerived):
            operand = _evaluate(node.operand, env)
            fn = _resolve_ibeam(node.dfn.path)
            return _call_ibeam(fn, operand)
        if isinstance(node.dfn, RankDerived):
            return _apply_rank_monadic(node.dfn, node.operand, env)
        dfn_val = _evaluate(node.dfn, env)
        operand = _evaluate(node.operand, env)
        if isinstance(dfn_val, IBeamDerived):
            fn = _resolve_ibeam(dfn_val.path)
            return _call_ibeam(fn, operand)
        if isinstance(dfn_val, FunctionRef):
            return _dispatch_monadic(dfn_val.glyph, operand, env)
        if not isinstance(dfn_val, _DfnClosure):
            raise DomainError(f"Expected dfn, got {type(dfn_val)}")
        return _call_dfn(dfn_val, operand)

    if isinstance(node, MonadicDopCall):
        dop_val = _evaluate(node.op_name, env)
        if not isinstance(dop_val, _DfnClosure):
            raise DomainError(f"Expected operator, got {type(dop_val)}")
        operand = _evaluate(node.operand, env)
        argument = _evaluate(node.argument, env)
        return _call_dfn(dop_val, argument, alpha_alpha=operand)

    if isinstance(node, DyadicDopCall):
        dop_val = _evaluate(node.op_name, env)
        if not isinstance(dop_val, _DfnClosure):
            raise DomainError(f"Expected operator, got {type(dop_val)}")
        left_operand = _evaluate(node.operand, env)   # ⍺⍺
        right_operand = _evaluate(node.left, env)      # ⍵⍵
        argument = _evaluate(node.right, env)           # ⍵
        return _call_dfn(dop_val, argument,
                         alpha_alpha=left_operand,
                         omega_omega=right_operand)

    if isinstance(node, DyadicDfnCall):
        if isinstance(node.dfn, SysVar):
            return _call_sys_function_dyadic(node.dfn.name, node.left, node.right, env)
        if isinstance(node.dfn, IBeamDerived):
            right = _evaluate(node.right, env)
            left = _evaluate(node.left, env)
            fn = _resolve_ibeam(node.dfn.path)
            return _call_ibeam_dyadic(fn, left, right)
        if isinstance(node.dfn, RankDerived):
            return _apply_rank_dyadic(node.dfn, node.left, node.right, env)
        dfn_val = _evaluate(node.dfn, env)
        right = _evaluate(node.right, env)
        left = _evaluate(node.left, env)
        if isinstance(dfn_val, IBeamDerived):
            fn = _resolve_ibeam(dfn_val.path)
            return _call_ibeam_dyadic(fn, left, right)
        if isinstance(dfn_val, FunctionRef):
            return _dispatch_dyadic(dfn_val.glyph, left, right, env)
        if not isinstance(dfn_val, _DfnClosure):
            raise DomainError(f"Expected dfn, got {type(dfn_val)}")
        return _call_dfn(dfn_val, right, alpha=left)

    if isinstance(node, Assignment):
        # Protect read-only quad-names
        _READONLY_QUADS = {"⎕A", "⎕D", "⎕TS", "⎕EN", "⎕DM"}
        if node.name in _READONLY_QUADS:
            raise DomainError(f"Cannot assign to read-only system variable {node.name}")
        value = _evaluate(node.value, env)
        # Classify and check name class
        if isinstance(value, (_DfnClosure, IBeamDerived)):
            new_class = NC_FUNCTION
        elif isinstance(value, APLArray):
            new_class = NC_ARRAY
        else:
            new_class = NC_UNKNOWN
        name_table = env.get("__name_table__", {})
        if node.name in name_table and name_table[node.name] != new_class:
            raise ClassError("Cannot change class of " + node.name + " from " + str(name_table[node.name]) + " to " + str(new_class))
        name_table[node.name] = new_class
        env["__name_table__"] = name_table
        env[node.name] = value
        if node.name == "⎕RL" and isinstance(value, APLArray):
            _random.seed(int(value.data[0]))
        if node.name == "⎕FR" and isinstance(value, APLArray):
            fr_val = int(value.data[0])
            if fr_val not in (645, 1287):
                raise DomainError("⎕FR must be 645 or 1287")
        return value if isinstance(value, APLArray) else S(0)

    if isinstance(node, InnerProduct):
        right = _evaluate(node.right, env)
        left = _evaluate(node.left, env)
        left_fn = _lookup_dyadic(node.left_fn)
        right_fn = _lookup_dyadic(node.right_fn)
        if left_fn is None or right_fn is None:
            raise DomainError(f"Unknown function in inner product")
        reduce_fn: Callable[[APLArray, APLArray], APLArray] = left_fn  # type: ignore[assignment]
        apply_fn: Callable[[APLArray, APLArray], APLArray] = right_fn  # type: ignore[assignment]
        return _inner_product(reduce_fn, apply_fn, left, right)

    if isinstance(node, OuterProduct):
        right = _evaluate(node.right, env)
        left = _evaluate(node.left, env)
        # Fast path: numpy ufunc.outer
        ufunc_name = _GLYPH_UFUNC.get(node.function)
        if (
            ufunc_name
            and is_numeric_array(left.data)
            and is_numeric_array(right.data)
        ):
            ufunc = getattr(np, ufunc_name, None)
            if ufunc is not None and hasattr(ufunc, "outer"):
                from marple.backend import _OVERFLOW_UFUNCS
                a = np.reshape(left.data, left.shape) if len(left.shape) > 1 else left.data
                b = np.reshape(right.data, right.shape) if len(right.shape) > 1 else right.data
                if ufunc_name in _OVERFLOW_UFUNCS:
                    a = maybe_upcast(a)
                    b = maybe_upcast(b)
                result = ufunc.outer(a, b)
                if hasattr(result, "shape") and len(result.shape) == 0:  # type: ignore[union-attr]
                    raw = result.item()  # type: ignore[union-attr]
                    if ufunc_name in _OVERFLOW_UFUNCS:
                        raw = maybe_downcast_scalar(raw, _DOWNCAST_CT)
                    return S(raw)
                result_shape = list(result.shape) if hasattr(result, "shape") else []  # type: ignore[union-attr]
                result_flat = result.ravel() if hasattr(result, "ravel") else result  # type: ignore[union-attr]
                if ufunc_name in _OVERFLOW_UFUNCS:
                    result_flat = maybe_downcast(result_flat, _DOWNCAST_CT)
                return APLArray(result_shape, result_flat)
        func = _lookup_dyadic(node.function)
        if func is None:
            raise DomainError(f"Unknown function in outer product: {node.function}")
        apply_fn_: Callable[[APLArray, APLArray], APLArray] = func  # type: ignore[assignment]
        return _outer_product(apply_fn_, left, right)

    if isinstance(node, DerivedFunc):
        operand = _evaluate(node.operand, env)
        # Resolve function — may be a glyph string, AST node, or ⍺⍺/⍵⍵
        func: Any = None
        if isinstance(node.function, str):
            func = _lookup_dyadic(node.function)
        elif isinstance(node.function, (AlphaAlpha, OmegaOmega)):
            fn_val = _evaluate(node.function, env)
            if isinstance(fn_val, FunctionRef):
                func = _lookup_dyadic(fn_val.glyph)
            elif isinstance(fn_val, _DfnClosure):
                func = lambda a, o, _c=fn_val: _call_dfn(_c, o, alpha=a)
            else:
                raise DomainError("⍺⍺ is not a function in this context")
        elif isinstance(node.function, (Dfn, Var)):
            fn_val = _evaluate(node.function, env)
            if isinstance(fn_val, _DfnClosure):
                func = lambda a, o, _c=fn_val: _call_dfn(_c, o, alpha=a)
            else:
                raise DomainError(f"Expected function for operator, got {type(fn_val)}")
        elif isinstance(node.function, FunctionRef):
            func = _lookup_dyadic(node.function.glyph)
        if func is None:
            raise DomainError(f"Unknown function for operator: {node.function}")
        if node.operator == "/":
            return _reduce(func, operand)  # type: ignore[arg-type]
        if node.operator == "⌿":
            return _reduce_first(func, operand)  # type: ignore[arg-type]
        if node.operator == "\\":
            return _scan(func, operand)  # type: ignore[arg-type]
        if node.operator == "⍀":
            return _scan_first(func, operand)  # type: ignore[arg-type]
        raise DomainError(f"Unknown operator: {node.operator}")

    if isinstance(node, Program):
        result: APLArray | _DfnClosure = S(0)
        for stmt in node.statements:
            result = _evaluate(stmt, env)
        if isinstance(result, APLArray):
            return result
        return S(0)

    raise DomainError(f"Unknown AST node: {type(node)}")


# Map glyphs to numpy ufunc names for fast outer product
_GLYPH_UFUNC: dict[str, str] = {
    "+": "add", "-": "subtract", "×": "multiply", "÷": "divide",
    "⌈": "maximum", "⌊": "minimum", "*": "power",
    "<": "less", "≤": "less_equal", "=": "equal",
    "≥": "greater_equal", ">": "greater", "≠": "not_equal",
}

# Map function objects to numpy ufunc names for fast reduce/scan
_UFUNC_MAP: dict[object, str] = {
    add: "add",
    subtract: "subtract",
    multiply: "multiply",
    divide: "divide",
    maximum: "maximum",
    minimum: "minimum",
    power: "power",
}

# Commutative ops where right-to-left = left-to-right
_COMMUTATIVE = {add, multiply, maximum, minimum}


def _reduce_vector(
    func: Callable[[APLArray, APLArray], APLArray],
    data: list[Any],
) -> Any:
    """Reduce a flat list right-to-left, returning a scalar value."""
    result = S(data[-1])
    for i in range(len(data) - 2, -1, -1):
        result = func(S(data[i]), result)
    return result.data[0]


def _reduce(
    func: Callable[[APLArray, APLArray], APLArray],
    omega: APLArray,
) -> APLArray:
    data = omega.data
    if len(data) == 0:
        raise DomainError("Cannot reduce empty array")

    # Fast path: use numpy ufunc.reduce (commutative ops only)
    if func in _COMMUTATIVE:
        ufunc_name = _UFUNC_MAP.get(func)
        if ufunc_name and is_numeric_array(data):
            ufunc = getattr(np, ufunc_name, None)
            if ufunc is not None and hasattr(ufunc, "reduce"):
                work_data = maybe_upcast(data)
                if len(omega.shape) <= 1:
                    raw = ufunc.reduce(work_data).item()
                    return S(maybe_downcast_scalar(raw, _DOWNCAST_CT))
                shaped = np.reshape(work_data, omega.shape)
                result = ufunc.reduce(shaped, axis=-1)
                result = maybe_downcast(result, _DOWNCAST_CT)
                return APLArray(list(result.shape), result.ravel())

    # Fallback: Python loop
    # Reduce along last axis
    if len(omega.shape) <= 1:
        return S(_reduce_vector(func, to_list(data)))
    # Higher rank: reduce each row (last-axis slice)
    last = omega.shape[-1]
    new_shape = omega.shape[:-1]
    data_list = to_list(data)
    n_rows = len(data_list) // last
    results: list[Any] = []
    for i in range(n_rows):
        row = data_list[i * last : (i + 1) * last]
        results.append(_reduce_vector(func, row))
    return APLArray(new_shape, results)


def _scan(
    func: Callable[[APLArray, APLArray], APLArray],
    omega: APLArray,
) -> APLArray:
    if len(omega.shape) <= 1:
        data = omega.data
        if len(data) == 0:
            return APLArray([0], [])
        results = [data[0]]
        acc = S(data[0])
        for i in range(1, len(data)):
            acc = func(acc, S(data[i]))
            results.append(acc.data[0])
        return APLArray([len(results)], results)
    # Higher rank: scan along last axis
    last = omega.shape[-1]
    n_rows = len(omega.data) // last
    results: list[Any] = []
    for i in range(n_rows):
        row = omega.data[i * last : (i + 1) * last]
        acc = S(row[0])
        results.append(row[0])
        for j in range(1, last):
            acc = func(acc, S(row[j]))
            results.append(acc.data[0])
    return APLArray(list(omega.shape), results)


def _reduce_first(
    func: Callable[[APLArray, APLArray], APLArray],
    omega: APLArray,
) -> APLArray:
    """Reduce along the first axis (⌿)."""
    if len(omega.shape) <= 1:
        return _reduce(func, omega)
    # Extract major cells and reduce pairwise
    first = omega.shape[0]
    cell_shape = omega.shape[1:]
    cell_size = 1
    for s in cell_shape:
        cell_size *= s
    data_list = to_list(omega.data)
    # Start with first major cell
    acc = data_list[:cell_size]
    for i in range(1, first):
        cell = data_list[i * cell_size : (i + 1) * cell_size]
        paired = []
        for a, b in zip(acc, cell):
            paired.append(func(S(a), S(b)).data[0])
        acc = paired
    return APLArray(cell_shape, acc)


def _scan_first(
    func: Callable[[APLArray, APLArray], APLArray],
    omega: APLArray,
) -> APLArray:
    """Scan along the first axis (⍀)."""
    if len(omega.shape) <= 1:
        return _scan(func, omega)
    first = omega.shape[0]
    cell_shape = omega.shape[1:]
    cell_size = 1
    for s in cell_shape:
        cell_size *= s
    data_list = to_list(omega.data)
    # First major cell is unchanged
    acc = data_list[:cell_size]
    results = list(acc)
    for i in range(1, first):
        cell = data_list[i * cell_size : (i + 1) * cell_size]
        new_acc = []
        for a, b in zip(acc, cell):
            new_acc.append(func(S(a), S(b)).data[0])
        acc = new_acc
        results.extend(acc)
    return APLArray(list(omega.shape), results)


def _inner_product(
    reduce_fn: Callable[[APLArray, APLArray], APLArray],
    apply_fn: Callable[[APLArray, APLArray], APLArray],
    alpha: APLArray,
    omega: APLArray,
) -> APLArray:
    # Fast path: +.× is matrix multiply — use np.tensordot
    # (np.dot uses wrong axis pairing for rank > 2)
    if (
        reduce_fn is add
        and apply_fn is multiply
        and is_numeric_array(alpha.data)
        and is_numeric_array(omega.data)
        and hasattr(np, "tensordot")
    ):
        # Check compatible dimensions: last axis of A must match first axis of B
        if len(alpha.shape) <= 1 and len(omega.shape) <= 1:
            if len(alpha.data) != len(omega.data):
                raise LengthError(f"Inner product length error: {len(alpha.data)} vs {len(omega.data)}")
        a = np.reshape(alpha.data, alpha.shape) if len(alpha.shape) > 1 else alpha.data
        b = np.reshape(omega.data, omega.shape) if len(omega.shape) > 1 else omega.data
        a = maybe_upcast(a)
        b = maybe_upcast(b)
        result = np.tensordot(a, b, axes=([-1], [0]))
        if hasattr(result, "shape") and len(result.shape) == 0:
            return S(maybe_downcast_scalar(result.item(), _DOWNCAST_CT))
        result_shape = list(result.shape) if hasattr(result, "shape") else []
        result_flat = result.ravel() if hasattr(result, "ravel") else result
        if is_numeric_array(result_flat):
            result_flat = maybe_downcast(result_flat, _DOWNCAST_CT)
        return APLArray(result_shape, result_flat)

    # Vector inner product: reduce(apply(a, b))
    if len(alpha.shape) <= 1 and len(omega.shape) <= 1:
        if len(alpha.data) != len(omega.data):
            raise LengthError(f"Inner product length error: {len(alpha.data)} vs {len(omega.data)}")
        paired = [apply_fn(S(a), S(b)) for a, b in zip(alpha.data, omega.data)]
        result = paired[-1]
        for i in range(len(paired) - 2, -1, -1):
            result = reduce_fn(paired[i], result)
        return result
    # Matrix inner product
    if len(alpha.shape) == 2 and len(omega.shape) == 2:
        m, k1 = alpha.shape
        k2, n = omega.shape
        if k1 != k2:
            raise LengthError(f"Inner product shape mismatch: {alpha.shape} vs {omega.shape}")
        result_data: list[object] = []
        for i in range(m):
            for j in range(n):
                row = [alpha.data[i * k1 + p] for p in range(k1)]
                col = [omega.data[p * n + j] for p in range(k2)]
                paired = [apply_fn(S(a), S(b)) for a, b in zip(row, col)]
                val = paired[-1]
                for idx in range(len(paired) - 2, -1, -1):
                    val = reduce_fn(paired[idx], val)
                result_data.append(val.data[0])
        return APLArray([m, n], result_data)
    # Vector × matrix
    if len(alpha.shape) == 1 and len(omega.shape) == 2:
        k, n = omega.shape
        if len(alpha.data) != k:
            raise LengthError(f"Inner product shape mismatch")
        result_data = []
        for j in range(n):
            col = [omega.data[p * n + j] for p in range(k)]
            paired = [apply_fn(S(a), S(b)) for a, b in zip(alpha.data, col)]
            val = paired[-1]
            for idx in range(len(paired) - 2, -1, -1):
                val = reduce_fn(paired[idx], val)
            result_data.append(val.data[0])
        return APLArray([n], result_data)
    # Matrix × vector
    if len(alpha.shape) == 2 and len(omega.shape) == 1:
        m, k1 = alpha.shape
        if k1 != len(omega.data):
            raise LengthError(f"Inner product shape mismatch")
        a_data = to_list(alpha.data)
        b_data = to_list(omega.data)
        result_data = []
        for i in range(m):
            row = [a_data[i * k1 + p] for p in range(k1)]
            paired = [apply_fn(S(a), S(b)) for a, b in zip(row, b_data)]
            val = paired[-1]
            for idx in range(len(paired) - 2, -1, -1):
                val = reduce_fn(paired[idx], val)
            result_data.append(val.data[0])
        return APLArray([m], result_data)
    raise RankError(f"Inner product not supported for shapes {alpha.shape} and {omega.shape}")


def _outer_product(
    func: Callable[[APLArray, APLArray], APLArray],
    alpha: APLArray,
    omega: APLArray,
) -> APLArray:
    result_data: list[object] = []
    for a in alpha.data:
        for b in omega.data:
            result_data.append(func(S(a), S(b)).data[0])
    result_shape = alpha.shape + omega.shape
    if not result_shape:
        return S(result_data[0])
    return APLArray(result_shape, result_data)


_CT_COMPARISONS: dict[str, object] = {
    "<": less_than,
    "≤": less_equal,
    "=": equal,
    "≥": greater_equal,
    ">": greater_than,
    "≠": not_equal,
}

def _lookup_dyadic(name: str) -> object | None:
    """Look up a dyadic function by glyph, including CT comparisons."""
    f = DYADIC_FUNCTIONS.get(name)
    if f is not None:
        return f
    return _CT_COMPARISONS.get(name)


_sys_workspace: Namespace | None = None


def _get_sys_workspace() -> Namespace:
    global _sys_workspace
    if _sys_workspace is None:
        import marple.stdlib
        f = marple.stdlib.__file__
        # Get directory: strip filename from path
        stdlib_path = f[:f.rfind("/")]  if "/" in f else f[:f.rfind("\\")]
        _sys_workspace = load_system_workspace(stdlib_path)
    return _sys_workspace


def _resolve_qualified(parts: list[str], env: dict[str, Any]) -> Any:
    """Resolve a qualified name like ['$', 'str', 'upper']."""
    if parts[0] == "$":
        sys_ws = _get_sys_workspace()
        result = sys_ws.resolve(parts[1:])
        if result is None:
            raise ValueError_("Undefined: " + "::".join(parts))
        return result  # type: ignore[return-value]
    # User namespace — look in env namespaces
    ns = env.get("__namespaces__", {})
    if parts[0] in ns:
        result = ns[parts[0]].resolve(parts[1:])
        if result is None:
            raise ValueError_("Undefined: " + "::".join(parts))
        return result  # type: ignore[return-value]
    raise ValueError_(f"Undefined namespace: {parts[0]}")


_ibeam_cache: dict[str, Any] = {}


def _resolve_ibeam(path: str) -> Any:
    """Resolve an i-beam path string to a Python callable."""
    if path in _ibeam_cache:
        return _ibeam_cache[path]
    # Check allowlist
    try:
        allowed = os.environ.get("MARPLE_IBEAM_ALLOW")
    except AttributeError:
        allowed = None
    if allowed is not None:
        prefixes = [p.strip() for p in allowed.split(",")]
        if not any(path.startswith(p) for p in prefixes):
            raise SecurityError(f"i-beam path not allowed: {path}")
    # Resolve
    parts = path.rsplit(".", 1)
    if len(parts) != 2:
        raise DomainError(f"invalid i-beam path: {path}")
    module_path, func_name = parts
    try:
        module = __import__(module_path)
        # __import__ returns the top-level module; traverse to the leaf
        for part in module_path.split(".")[1:]:
            module = getattr(module, part)
        fn = getattr(module, func_name)
    except (ImportError, AttributeError) as e:
        raise DomainError(str(e)) from e
    _ibeam_cache[path] = fn
    return fn


def _call_ibeam(fn: Any, right: APLArray) -> APLArray:
    """Call a Python function monadically via i-beam."""
    try:
        result = fn(right)
    except Exception as e:
        raise DomainError(f"{e}") from e
    if not isinstance(result, APLArray):
        raise DomainError(f"i-beam function must return APLArray, got {type(result)}")
    return result


def _call_ibeam_dyadic(fn: Any, left: APLArray, right: APLArray) -> APLArray:
    """Call a Python function dyadically via i-beam."""
    try:
        result = fn(left, right)
    except Exception as e:
        raise DomainError(f"{e}") from e
    if not isinstance(result, APLArray):
        raise DomainError(f"i-beam function must return APLArray, got {type(result)}")
    return result


def _get_io(env: dict[str, Any]) -> int:
    return int(env["⎕IO"].data[0])


def _get_ct(env: dict[str, Any]) -> float:
    return float(env["⎕CT"].data[0])


# ── Monadic handlers: (operand, env) → APLArray ──

def _m_execute(operand: APLArray, env: dict[str, Any]) -> APLArray:
    source = "".join(str(c) for c in operand.data)
    return _evaluate(parse(source, env.get("__name_table__")), env)


def _m_format(operand: APLArray, env: dict[str, Any]) -> APLArray:
    return _format_array(operand)


def _m_iota(operand: APLArray, env: dict[str, Any]) -> APLArray:
    io = _get_io(env)
    n = int(operand.data[0])
    return APLArray([n], list(range(io, n + io)))


def _m_tally(operand: APLArray, env: dict[str, Any]) -> APLArray:
    return S(1) if operand.is_scalar() else S(operand.shape[0])


def _m_grade_up(operand: APLArray, env: dict[str, Any]) -> APLArray:
    return grade_up(operand, _get_io(env))


def _m_grade_down(operand: APLArray, env: dict[str, Any]) -> APLArray:
    return grade_down(operand, _get_io(env))


def _roll(omega: APLArray, env: dict[str, Any]) -> APLArray:
    """Monadic ?: roll. ?N → random int ⎕IO..N, ?0 → random float [0,1)."""
    io = _get_io(env)
    data = to_list(omega.data)
    results: list[object] = []
    for x in data:
        n = int(x)
        if n == 0:
            results.append(_random.random())
        else:
            results.append(_random.randint(io, n - 1 + io))
    if omega.is_scalar():
        return S(results[0])
    return APLArray(list(omega.shape), results)


def _deal(left: APLArray, right: APLArray, env: dict[str, Any]) -> APLArray:
    """Dyadic ?: deal. N?M → N distinct random integers from ⎕IO..M."""
    io = _get_io(env)
    n = int(left.data[0])
    m = int(right.data[0])
    if n > m:
        raise LengthError(f"Deal: cannot choose {n} from {m}")
    result = _random.sample(range(io, m + io), n)
    return APLArray([n], result)


_MONADIC_DISPATCH: dict[str, Callable[[APLArray, dict[str, Any]], APLArray]] = {
    "⍎": _m_execute,
    "⍕": _m_format,
    "⍳": _m_iota,
    "≢": _m_tally,
    "⍋": _m_grade_up,
    "⍒": _m_grade_down,
    "?": _roll,
}


# ── Dyadic handlers: (left, right, env) → APLArray ──

def _d_format(left: APLArray, right: APLArray, env: dict[str, Any]) -> APLArray:
    return _dyadic_format(left, right)


def _d_index_of(left: APLArray, right: APLArray, env: dict[str, Any]) -> APLArray:
    return index_of(left, right, _get_io(env), _get_ct(env))


def _d_membership(left: APLArray, right: APLArray, env: dict[str, Any]) -> APLArray:
    return membership(left, right, _get_ct(env))


def _d_from(left: APLArray, right: APLArray, env: dict[str, Any]) -> APLArray:
    return from_array(left, right, _get_io(env))


def _d_match(left: APLArray, right: APLArray, env: dict[str, Any]) -> APLArray:
    return S(1 if left == right else 0)


def _d_not_match(left: APLArray, right: APLArray, env: dict[str, Any]) -> APLArray:
    return S(0 if left == right else 1)


def _d_ct_compare(left: APLArray, right: APLArray, env: dict[str, Any], glyph: str) -> APLArray:
    return _CT_COMPARISONS[glyph](left, right, _get_ct(env))  # type: ignore[operator]


_DYADIC_DISPATCH: dict[str, Callable[[APLArray, APLArray, dict[str, Any]], APLArray]] = {
    "⍕": _d_format,
    "⍳": _d_index_of,
    "∈": _d_membership,
    "?": _deal,
    "⌷": _d_from,
    "≡": _d_match,
    "≢": _d_not_match,
}


def _dispatch_monadic(glyph: str, operand: APLArray, env: dict[str, Any]) -> APLArray:
    """Dispatch a monadic primitive function."""
    handler = _MONADIC_DISPATCH.get(glyph)
    if handler is not None:
        return handler(operand, env)
    func = MONADIC_FUNCTIONS.get(glyph)
    if func is not None:
        return func(operand)  # type: ignore[operator]
    raise DomainError(f"Unknown monadic function: {glyph}")


def _dispatch_dyadic(glyph: str, left: APLArray, right: APLArray, env: dict[str, Any]) -> APLArray:
    """Dispatch a dyadic primitive function."""
    handler = _DYADIC_DISPATCH.get(glyph)
    if handler is not None:
        return handler(left, right, env)
    if glyph in _CT_COMPARISONS:
        return _d_ct_compare(left, right, env, glyph)
    func = _lookup_dyadic(glyph)
    if func is not None:
        return func(left, right)  # type: ignore[operator]
    raise DomainError(f"Unknown dyadic function: {glyph}")


def _monadic_dr(operand: APLArray) -> APLArray:
    """Monadic ⎕DR: query internal data representation."""
    from marple.backend import _is_int_dtype, _is_float_dtype
    # Character data
    if len(operand.data) > 0 and isinstance(to_list(operand.data)[0], str):
        return S(80)
    # Numeric array with dtype
    if is_numeric_array(operand.data):
        dtype_str = str(operand.data.dtype)
        if "uint8" in dtype_str:
            return S(11)  # boolean
        if "int8" in dtype_str:
            return S(83)
        if "int16" in dtype_str:
            return S(163)
        if "int32" in dtype_str:
            return S(323)
        if "int64" in dtype_str:
            return S(645)  # treat as float-equivalent width
        if _is_float_dtype(operand.data):
            return S(645)
    # Python list fallback
    vals = to_list(operand.data)
    if vals:
        v = vals[0]
        if isinstance(v, str):
            return S(80)
        if isinstance(v, float):
            return S(645)
        if isinstance(v, int):
            return S(323)
    return S(0)


def _dyadic_dr(left: APLArray, right: APLArray) -> APLArray:
    """Dyadic ⎕DR: convert data representation."""
    target = int(left.data[0])
    vals = to_list(right.data)
    if target == 645:
        # Convert to float
        new_data = [float(v) for v in vals]
        return APLArray(list(right.shape), new_data)
    if target in (323, 163, 83):
        # Convert to integer
        new_data = [int(round(v)) for v in vals]
        return APLArray(list(right.shape), new_data)
    if target == 11:
        # Convert to boolean (uint8)
        from marple.backend import to_bool_array
        new_data = to_bool_array([int(bool(v)) for v in vals])
        return APLArray(list(right.shape), new_data)
    if target == 80:
        # Convert to character (via UCS)
        new_data = [chr(int(v)) for v in vals]
        return APLArray(list(right.shape), new_data)
    raise DomainError("Invalid ⎕DR type code: " + str(target))


def _call_sys_function_monadic(name: str, operand_node: object, env: dict[str, Any]) -> APLArray:
    """Dispatch a monadic system function call."""
    operand = _evaluate(operand_node, env)
    if name == "⎕UCS":
        if all(isinstance(x, str) for x in operand.data):
            # Chars to codepoints
            return APLArray(list(operand.shape), [ord(c) for c in operand.data])
        else:
            # Codepoints to chars
            data = to_list(operand.data)
            if operand.is_scalar():
                return APLArray([], [chr(int(data[0]))])
            return APLArray(list(operand.shape), [chr(int(x)) for x in data])
    if name == "⎕NC":
        # Expect a character vector (name)
        nc_name = "".join(str(c) for c in operand.data)
        name_table = env.get("__name_table__", {})
        return S(name_table.get(nc_name, 0))
    if name == "⎕EX":
        ex_name = "".join(str(c) for c in operand.data)
        name_table = env.get("__name_table__", {})
        if ex_name in env:
            del env[ex_name]
            if ex_name in name_table:
                del name_table[ex_name]
            return S(1)
        return S(0)
    if name == "⎕DR":
        return _monadic_dr(operand)
    if name == "⎕SIGNAL":
        code = int(operand.data[0])
        _ERROR_MAP = {
            1: SyntaxError_, 2: ValueError_, 3: DomainError, 4: LengthError,
            5: RankError, 6: IndexError_, 7: LimitError, 9: SecurityError,
        }
        err_class = _ERROR_MAP.get(code, DomainError)
        raise err_class(f"Signalled by ⎕SIGNAL {code}")
    if name == "⎕NREAD":
        import os as _os
        path = "".join(str(c) for c in operand.data)
        with open(path) as f:
            text = f.read()
        chars = list(text)
        return APLArray([len(chars)], chars) if chars else APLArray([0], [])
    if name == "⎕NEXISTS":
        import os as _os
        path = "".join(str(c) for c in operand.data)
        try:
            _os.stat(path)
            return S(1)
        except OSError:
            return S(0)
    if name == "⎕NDELETE":
        import os as _os
        path = "".join(str(c) for c in operand.data)
        try:
            _os.remove(path)
        except OSError:
            raise DomainError("File not found: " + path)
        return S(0)
    if name == "⎕CR":
        fn_name = "".join(str(c) for c in operand.data)
        sources = env.get("__sources__", {})
        if fn_name not in sources:
            raise DomainError("Not a defined function: " + fn_name)
        source = sources[fn_name]
        # Source may be a list of lines (from matrix ⎕FX) or a string
        if isinstance(source, list):
            lines = source
        else:
            lines = [source]
        max_len = max(len(l) for l in lines) if lines else 0
        flat: list[object] = []
        for line in lines:
            flat.extend(list(line.ljust(max_len)))
        return APLArray([len(lines), max_len], flat)
    if name == "⎕FX":
        # Accept vector (single line) or matrix (one row per line)
        if len(operand.shape) == 2:
            # Matrix: extract rows, strip trailing spaces, join with newlines
            rows, cols = operand.shape
            lines = []
            for r in range(rows):
                row_chars = operand.data[r * cols : (r + 1) * cols]
                lines.append("".join(str(c) for c in row_chars).rstrip())
            text = "\n".join(lines)
        else:
            text = "".join(str(c) for c in operand.data)
        parts = text.split("←", 1)
        if len(parts) < 2:
            raise DomainError("⎕FX requires an assignment: name←{body}")
        fn_name = parts[0].strip()
        try:
            interpret(text, env)
        except APLError:
            raise DomainError("⎕FX: invalid function definition")
        if fn_name not in env or not isinstance(env[fn_name], _DfnClosure):
            raise DomainError("⎕FX did not produce a function")
        # Store source as lines for matrix round-trip
        if len(operand.shape) == 2:
            sources = env.get("__sources__", {})
            sources[fn_name] = [
                "".join(str(c) for c in operand.data[r * cols : (r + 1) * cols]).rstrip()
                for r in range(rows)
            ]
            env["__sources__"] = sources
        chars = list(fn_name)
        return APLArray([len(chars)], chars)
    if name == "⎕FMT":
        if isinstance(operand_node, FmtArgs):
            values = [_evaluate(arg, env) for arg in operand_node.args]
            # Monadic with multiple args: format each, join
            from marple.formatting import format_num
            parts = []
            for v in values:
                parts.append(_monadic_fmt(v))
            # Concatenate the character vectors
            all_chars: list[object] = []
            for p in parts:
                all_chars.extend(p.data)
                all_chars.append(" ")
            if all_chars:
                all_chars.pop()  # remove trailing space
            return APLArray([len(all_chars)], all_chars)
        operand = _evaluate(operand_node, env)
        return _monadic_fmt(operand)
    raise DomainError(f"Unknown system function: {name}")


def _monadic_fmt(operand: APLArray) -> APLArray:
    """Monadic ⎕FMT: default formatting as character vector."""
    from marple.formatting import format_num
    if operand.shape == []:
        # Scalar
        text = format_num(operand.data[0])
    elif len(operand.shape) == 1:
        # Vector
        if len(operand.data) > 0 and isinstance(operand.data[0], str):
            text = "".join(str(c) for c in operand.data)
        else:
            text = " ".join(format_num(x) for x in operand.data)
    else:
        text = str(operand)
    return APLArray([len(text)], list(text))


class _FmtGroup:
    """A format group: one data column's format, or a text insertion.

    For data groups: code is I/F/E/A/G, repeat is how many times to apply
    (e.g. 5A1 → code='A', width=1, repeat=5). Total chars = repeat * width.
    For text: code is 'TEXT', text holds the literal.
    """
    def __init__(self, code: str, width: int = 0, decimals: int = 0,
                 repeat: int = 1, text: str = "") -> None:
        self.code = code
        self.width = width
        self.decimals = decimals
        self.repeat = repeat
        self.text = text


def _parse_fmt_spec(spec: str) -> list[_FmtGroup]:
    """Parse a format specification string like 'I5,F8.2,⊂ => ⊃,5A1'.

    Each format code (with optional repetition) is ONE group consuming
    ONE data column. Text insertions don't consume columns.
    Supports both ⊂text⊃ and <text> delimiters.
    """
    groups: list[_FmtGroup] = []
    i = 0
    while i < len(spec):
        if spec[i] in " ,":
            i += 1
            continue
        # Text insertion: ⊂text⊃ or <text>
        if spec[i] in ("⊂", "<"):
            close = "⊃" if spec[i] == "⊂" else ">"
            i += 1
            end = spec.index(close, i)
            groups.append(_FmtGroup("TEXT", text=spec[i:end]))
            i = end + 1
            continue
        # Optional repetition count
        rep = 0
        while i < len(spec) and spec[i].isdigit():
            rep = rep * 10 + int(spec[i])
            i += 1
        if i >= len(spec):
            break
        code = spec[i].upper()
        if code not in "IFEAG":
            raise DomainError(f"Unknown format code: {spec[i]}")
        i += 1
        # G format with pattern: G⊂pattern⊃ or G<pattern>
        if code == "G" and i < len(spec) and spec[i] in ("⊂", "<"):
            close = "⊃" if spec[i] == "⊂" else ">"
            i += 1
            end = spec.index(close, i)
            pattern = spec[i:end]
            i = end + 1
            groups.append(_FmtGroup("G", text=pattern, repeat=max(rep, 1)))
            continue
        # Width
        width = 0
        while i < len(spec) and spec[i].isdigit():
            width = width * 10 + int(spec[i])
            i += 1
        # Decimals
        decimals = 0
        if i < len(spec) and spec[i] == ".":
            i += 1
            while i < len(spec) and spec[i].isdigit():
                decimals = decimals * 10 + int(spec[i])
                i += 1
        # Validate: F format requires d <= w-2, E format requires d <= w-2
        if code == "F" and width > 0 and decimals > width - 2:
            raise DomainError(f"F format: decimals ({decimals}) > width-2 ({width - 2})")
        if code == "E" and width > 0 and decimals > width - 2:
            raise DomainError(f"E format: decimals ({decimals}) > width-2 ({width - 2})")
        groups.append(_FmtGroup(code, width, decimals, max(rep, 1)))
    return groups


def _format_one_value(code: str, width: int, decimals: int,
                      value: object) -> str:
    """Format a single value with a format code."""
    if value is None:
        return " " * width if width > 0 else ""
    if code == "A":
        text = value if isinstance(value, str) else str(value)
        return text.ljust(width) if width > 0 else text
    num = float(value) if not isinstance(value, (int, float)) else value
    if code == "I":
        text = str(int(num))
    elif code == "F":
        text = f"{num:.{decimals}f}"
    elif code == "E":
        text = f"{num:.{decimals}E}"
    elif code == "G":
        from marple.formatting import format_num
        text = format_num(num)
    else:
        text = str(num)
    text = text.replace("-", "¯")
    return text.rjust(width) if width > 0 else text


def _apply_g_pattern(pattern: str, value: object) -> str:
    """Apply a G-format pattern. '9' placeholders are filled with digits,
    other characters pass through literally."""
    num = abs(int(float(value) if not isinstance(value, (int, float)) else value))
    digit_count = pattern.count("9")
    digits = str(num).zfill(digit_count)
    result: list[str] = []
    di = 0
    for ch in pattern:
        if ch == "9":
            result.append(digits[di] if di < len(digits) else "0")
            di += 1
        else:
            result.append(ch)
    return "".join(result)


def _apply_group(group: _FmtGroup, value: APLArray | None,
                 row: int) -> str:
    """Apply a format group to one row of a column's data."""
    if group.code == "TEXT":
        return group.text
    if value is None:
        return " " * (group.width * group.repeat)

    # Type checking: A format requires character data, others require numeric
    if value is not None:
        is_char = (len(value.data) > 0 and isinstance(value.data[0], str))
        if group.code == "A" and not is_char:
            raise DomainError("A format requires character data")
        if group.code != "A" and is_char:
            raise DomainError(f"{group.code} format requires numeric data, got character")

    if group.code == "A":
        # Character data: extract repeat*width chars from row
        total_chars = group.repeat * group.width
        if len(value.shape) >= 2:
            # Matrix: extract chars from this row
            cols = value.shape[1]
            start = row * cols
            if start >= len(value.data):
                return " " * total_chars
            row_chars = [str(c) for c in value.data[start:start + cols]]
        elif len(value.data) > 0 and isinstance(value.data[0], str):
            # Character vector: one string, only valid for row 0
            row_chars = [str(c) for c in value.data] if row == 0 else []
        else:
            row_chars = []
        # Pad or truncate to total_chars
        while len(row_chars) < total_chars:
            row_chars.append(" ")
        # Format each repeated sub-field
        parts: list[str] = []
        for r in range(group.repeat):
            ch = row_chars[r * group.width:(r + 1) * group.width]
            field = "".join(ch)
            parts.append(field.ljust(group.width) if group.width > 0 else field)
        return "".join(parts)
    else:
        # Numeric: extract scalar for this row
        if len(value.shape) == 0:
            scalar = value.data[0] if row == 0 else None
        elif row < value.shape[0]:
            scalar = value.data[row]
        else:
            scalar = None
        # G format with pattern: replace 9s with digits
        if group.code == "G" and group.text:
            if scalar is None:
                return " " * len(group.text)
            return _apply_g_pattern(group.text, scalar)
        return _format_one_value(group.code, group.width, group.decimals,
                                scalar)


def _column_row_count(value: APLArray, is_alpha: bool) -> int:
    """How many output rows does a column argument produce?"""
    if len(value.shape) == 0:
        return 1
    if len(value.shape) >= 2:
        return value.shape[0]
    if is_alpha and len(value.data) > 0 and isinstance(value.data[0], str):
        return 1  # character vector = one string
    return value.shape[0]


def _dyadic_fmt(fmt_str: str, values: list[APLArray]) -> APLArray:
    """Dyadic ⎕FMT: format values according to specification.

    Each format group (e.g. I5, 5A1, ⊂text⊃) is one unit. Data groups
    consume one value column each. Text groups insert literals.
    Groups cycle to cover all value columns.
    """
    groups = _parse_fmt_spec(fmt_str)
    if not groups:
        raise DomainError("Empty format specification")

    # Build a template: match groups to value columns, cycling groups.
    # Each entry: (group, value_index_or_None)
    template: list[tuple[_FmtGroup, int | None]] = []
    group_idx = 0
    val_idx = 0
    while val_idx < len(values):
        g = groups[group_idx % len(groups)]
        group_idx += 1
        if g.code == "TEXT":
            template.append((g, None))
            continue
        template.append((g, val_idx))
        val_idx += 1
    # Add any trailing text groups from the cycle
    while True:
        g = groups[group_idx % len(groups)]
        if g.code != "TEXT":
            break
        template.append((g, None))
        group_idx += 1

    # Determine number of rows
    num_rows = 1
    for g, vi in template:
        if vi is not None:
            num_rows = max(num_rows,
                           _column_row_count(values[vi], g.code == "A"))

    # Build each row
    rows: list[str] = []
    for row_idx in range(num_rows):
        parts: list[str] = []
        for g, vi in template:
            v = values[vi] if vi is not None else None
            parts.append(_apply_group(g, v, row_idx))
        rows.append("".join(parts))

    # Pad to equal width for character matrix
    max_width = max(len(r) for r in rows) if rows else 0
    all_chars: list[object] = []
    for r in rows:
        all_chars.extend(list(r.ljust(max_width)))
    return APLArray([len(rows), max_width], all_chars)


def _call_sys_function_dyadic(name: str, left_node: object, right_node: object, env: dict[str, Any]) -> APLArray:
    """Dispatch a dyadic system function call."""
    if name == "⎕EA":
        right = _evaluate(right_node, env)
        left = _evaluate(left_node, env)
        right_str = "".join(str(c) for c in right.data)
        try:
            return interpret(right_str, env)
        except APLError as e:
            env["⎕EN"] = S(e.code)
            env["⎕DM"] = APLArray([len(str(e))], list(str(e)))
            left_str = "".join(str(c) for c in left.data)
            return interpret(left_str, env)
    if name == "⎕DR":
        left = _evaluate(left_node, env)
        right = _evaluate(right_node, env)
        return _dyadic_dr(left, right)
    if name == "⎕NWRITE":
        left = _evaluate(left_node, env)
        right = _evaluate(right_node, env)
        path = "".join(str(c) for c in right.data)
        text = "".join(str(c) for c in left.data)
        with open(path, "w") as f:
            f.write(text)
        return APLArray([0], [])
    if name == "⎕FMT":
        left = _evaluate(left_node, env)
        fmt_str = "".join(str(c) for c in left.data)
        # Right operand may be FmtArgs (semicolon list) or a single value
        if isinstance(right_node, FmtArgs):
            values = [_evaluate(arg, env) for arg in right_node.args]
        else:
            right = _evaluate(right_node, env)
            values = [right]
        return _dyadic_fmt(fmt_str, values)
    raise DomainError(f"Unknown dyadic system function: {name}")


def _apply_func_monadic(
    func: object,
    omega: APLArray,
    env: dict[str, Any],
) -> APLArray:
    """Apply a function monadically. Used by rank operator."""
    if isinstance(func, str):
        return _dispatch_monadic(func, omega, env)
    if isinstance(func, ReduceOp):
        f = _lookup_dyadic(func.function)
        if f is None:
            raise DomainError(f"Unknown function for reduce: {func.function}")
        return _reduce(f, omega)  # type: ignore[arg-type]
    if isinstance(func, ScanOp):
        f = _lookup_dyadic(func.function)
        if f is None:
            raise DomainError(f"Unknown function for scan: {func.function}")
        return _scan(f, omega)  # type: ignore[arg-type]
    if isinstance(func, (Dfn, Var)):
        val = _evaluate(func, env)
        if isinstance(val, _DfnClosure):
            return _call_dfn(val, omega)
        raise DomainError(f"Expected dfn, got {type(val)}")
    if isinstance(func, FunctionRef):
        return _dispatch_monadic(func.glyph, omega, env)
    if isinstance(func, BoundOperator):
        # A derived verb from operator binding, used inside rank
        node = func  # type: ignore[assignment]
        if isinstance(node.operator, str) and node.operator in ("/", "⌿"):
            inner = _lookup_dyadic(node.left_operand) if isinstance(node.left_operand, str) else None
            if inner and node.operator == "/":
                return _reduce(inner, omega)  # type: ignore[arg-type]
            if inner and node.operator == "⌿":
                return _reduce_first(inner, omega)  # type: ignore[arg-type]
        if isinstance(node.operator, str) and node.operator in ("\\", "⍀"):
            inner = _lookup_dyadic(node.left_operand) if isinstance(node.left_operand, str) else None
            if inner and node.operator == "\\":
                return _scan(inner, omega)  # type: ignore[arg-type]
            if inner and node.operator == "⍀":
                return _scan_first(inner, omega)  # type: ignore[arg-type]
    raise DomainError(f"Cannot apply as monadic function: {type(func)}")


def _apply_func_dyadic(
    func: object,
    alpha: APLArray,
    omega: APLArray,
    env: dict[str, Any],
) -> APLArray:
    """Apply a function dyadically. Used by rank operator."""
    if isinstance(func, str):
        return _dispatch_dyadic(func, alpha, omega, env)
    if isinstance(func, (Dfn, Var)):
        val = _evaluate(func, env)
        if isinstance(val, _DfnClosure):
            return _call_dfn(val, omega, alpha=alpha)
        raise DomainError(f"Expected dfn, got {type(val)}")
    raise DomainError(f"Cannot apply as dyadic function: {type(func)}")


def _apply_rank_monadic(
    rank_node: RankDerived,
    operand_node: object,
    env: dict[str, Any],
) -> APLArray:
    omega = _evaluate(operand_node, env)
    rank_spec_val = _evaluate(rank_node.rank_spec, env)
    a, _, _ = resolve_rank_spec(rank_spec_val)
    k = clamp_rank(a, len(omega.shape))
    frame_shape, cells = decompose(omega, k)
    results = [_apply_func_monadic(rank_node.function, cell, env) for cell in cells]
    return reassemble(frame_shape, results)


def _apply_rank_dyadic(
    rank_node: RankDerived,
    left_node: object,
    right_node: object,
    env: dict[str, Any],
) -> APLArray:
    alpha = _evaluate(left_node, env)
    omega = _evaluate(right_node, env)
    rank_spec_val = _evaluate(rank_node.rank_spec, env)
    _, b_rank, c_rank = resolve_rank_spec(rank_spec_val)
    b = clamp_rank(b_rank, len(alpha.shape))
    c = clamp_rank(c_rank, len(omega.shape))
    left_frame, left_cells = decompose(alpha, b)
    right_frame, right_cells = decompose(omega, c)
    # Frame agreement
    if left_frame == right_frame:
        pairs = list(zip(left_cells, right_cells))
        frame = left_frame
    elif left_frame == []:
        pairs = [(left_cells[0], rc) for rc in right_cells]
        frame = right_frame
    elif right_frame == []:
        pairs = [(lc, right_cells[0]) for lc in left_cells]
        frame = left_frame
    else:
        raise LengthError(f"Frame mismatch: {left_frame} vs {right_frame}")
    results = [_apply_func_dyadic(rank_node.function, lc, rc, env) for lc, rc in pairs]
    return reassemble(frame, results)


def _bracket_index(
    array: APLArray,
    indices: list[object | None],
    env: dict[str, Any],
    io: int,
) -> APLArray:
    data = to_list(array.data)

    def product(*lists):
        """Simple replacement for itertools.product."""
        if not lists:
            yield ()
            return
        for item in lists[0]:
            for rest in product(*lists[1:]):
                yield (item,) + rest
    # Build 0-based index lists for each axis, tracking index shapes
    axis_indices: list[list[int]] = []
    idx_shapes: list[list[int]] = []  # shape of each index expression
    for axis, idx_node in enumerate(indices):
        if idx_node is None:
            axis_indices.append(list(range(array.shape[axis])))
            idx_shapes.append([array.shape[axis]])
        else:
            idx = _evaluate(idx_node, env)
            vals = to_list(idx.data) if not idx.is_scalar() else [idx.data[0]]
            axis_indices.append([int(v) - io for v in vals])
            idx_shapes.append(idx.shape if not idx.is_scalar() else [])
    # Pad missing axes with full range
    for axis in range(len(indices), len(array.shape)):
        axis_indices.append(list(range(array.shape[axis])))
        idx_shapes.append([array.shape[axis]])
    # Compute strides
    strides = [1] * len(array.shape)
    for i in range(len(array.shape) - 2, -1, -1):
        strides[i] = strides[i + 1] * array.shape[i + 1]
    # Gather results
    result: list[object] = []
    for combo in product(*axis_indices):
        flat = sum(i * s for i, s in zip(combo, strides))
        result.append(data[flat])
    # Result shape: concatenation of each index expression's shape
    # (scalar indices are dropped, matching APL semantics)
    result_shape: list[int] = []
    for s in idx_shapes:
        result_shape.extend(s)
    if not result_shape:
        return S(result[0])
    return APLArray(result_shape, result)


def _handle_import(source: str, env: dict[str, Any]) -> APLArray:
    """Handle #import directive."""
    parts = source.split()
    # #import qualified::name [as alias]
    if len(parts) < 2:
        raise DomainError("Invalid #import directive")
    qualified = parts[1]
    alias = None
    if len(parts) >= 4 and parts[2] == "as":
        alias = parts[3]
    name_parts = qualified.split("::")
    # Resolve the value
    if name_parts[0] == "$":
        sys_ws = _get_sys_workspace()
        result = sys_ws.resolve(name_parts[1:])
        if result is None:
            raise ValueError_(f"Undefined: {qualified}")
    else:
        raise ValueError_(f"Import from non-system namespace not yet supported: {qualified}")
    # Bind in env and classify
    bind_name = alias if alias else name_parts[-1]
    env[bind_name] = result
    name_table = env.get("__name_table__", {})
    if isinstance(result, (_DfnClosure, IBeamDerived)):
        name_table[bind_name] = NC_FUNCTION
    elif isinstance(result, APLArray):
        name_table[bind_name] = NC_ARRAY
    env["__name_table__"] = name_table
    return S(0)


_SYSTEM_DEFAULTS: dict[str, Any] = {
    "⎕IO": S(1),
    "⎕CT": S(1e-14),
    "⎕PP": S(10),
    "⎕EN": S(0),
    "⎕DM": APLArray([0], []),
    "⎕A": APLArray([26], list("ABCDEFGHIJKLMNOPQRSTUVWXYZ")),
    "⎕D": APLArray([10], list("0123456789")),
    "⎕WSID": APLArray([8], list("CLEAR WS")),
    "⎕RL": S(1),
    "⎕FR": S(645),
}


def default_env() -> dict[str, Any]:
    """Create a fresh environment with all system variable defaults."""
    env = dict(_SYSTEM_DEFAULTS)
    _random.seed(1)
    return env


def _newlines_to_diamonds(source: str) -> str:
    """Convert newlines to ⋄ (statement separator), preserving strings."""
    result: list[str] = []
    in_string = False
    for ch in source:
        if ch == "'" and not in_string:
            in_string = True
            result.append(ch)
        elif ch == "'" and in_string:
            in_string = False
            result.append(ch)
        elif ch == "\n" and not in_string:
            result.append("⋄")
        else:
            result.append(ch)
    return "".join(result)


def interpret(source: str, env: dict[str, Any] | None = None) -> APLArray:
    env = env or default_env()
    # Handle #import directives
    if source.strip().startswith("#import"):
        return _handle_import(source.strip(), env)
    name_table = env.get("__name_table__", {})
    # System functions are always classified as functions
    for qfn in ("⎕EA", "⎕UCS", "⎕NC", "⎕EX", "⎕SIGNAL", "⎕DR",
                 "⎕NREAD", "⎕NWRITE", "⎕NEXISTS", "⎕NDELETE",
                 "⎕CR", "⎕FX"):
        name_table[qfn] = NC_FUNCTION
    env["__name_table__"] = name_table
    op_arity = env.get("__operator_arity__", {})
    # Convert newlines to diamonds (statement separators) outside strings
    source = _newlines_to_diamonds(source)
    tree = parse(source, name_table, op_arity)
    result = _evaluate(tree, env)
    # Track source for dfn assignments so workspace save can reconstruct them
    if isinstance(tree, Assignment):
        value = env.get(tree.name)
        if isinstance(value, _DfnClosure):
            if "__sources__" not in env:
                env["__sources__"] = {}
            env["__sources__"][tree.name] = source.strip()
            # Fix name class and arity: dops use ⍺⍺ or ⍵⍵
            if "⍺⍺" in source or "⍵⍵" in source:
                name_table = env.get("__name_table__", {})
                name_table[tree.name] = NC_OPERATOR
                env["__name_table__"] = name_table
                arity = 2 if "⍵⍵" in source else 1
                op_ar = env.get("__operator_arity__", {})
                op_ar[tree.name] = arity
                env["__operator_arity__"] = op_ar
    if isinstance(result, _DfnClosure):
        return S(0)
    return result



from marple.arraymodel import APLArray, S
from marple.cells import clamp_rank, decompose, reassemble, resolve_rank_spec


class TestResolveRankSpec:
    def test_scalar(self) -> None:
        assert resolve_rank_spec(S(1)) == (1, 1, 1)

    def test_two_vector(self) -> None:
        assert resolve_rank_spec(APLArray([2], [0, 1])) == (1, 0, 1)

    def test_three_vector(self) -> None:
        assert resolve_rank_spec(APLArray([3], [2, 0, 1])) == (2, 0, 1)


class TestClampRank:
    def test_positive_within_range(self) -> None:
        assert clamp_rank(1, 3) == 1

    def test_positive_exceeds_rank(self) -> None:
        assert clamp_rank(5, 3) == 3

    def test_negative(self) -> None:
        # ¯1 on rank 3 → 3-1 = 2
        assert clamp_rank(-1, 3) == 2

    def test_negative_clamped(self) -> None:
        assert clamp_rank(-5, 3) == 0

    def test_zero(self) -> None:
        assert clamp_rank(0, 3) == 0


class TestDecompose:
    def test_vector_rank0(self) -> None:
        # Vector decomposed into scalars
        frame, cells = decompose(APLArray([3], [10, 20, 30]), 0)
        assert frame == [3]
        assert len(cells) == 3
        assert cells[0] == S(10)
        assert cells[2] == S(30)

    def test_vector_rank1(self) -> None:
        # Whole vector is one cell
        v = APLArray([3], [10, 20, 30])
        frame, cells = decompose(v, 1)
        assert frame == []
        assert len(cells) == 1
        assert cells[0] == v

    def test_matrix_rank1(self) -> None:
        # 3×4 matrix → 3 row vectors
        m = APLArray([3, 4], list(range(1, 13)))
        frame, cells = decompose(m, 1)
        assert frame == [3]
        assert len(cells) == 3
        assert cells[0] == APLArray([4], [1, 2, 3, 4])
        assert cells[2] == APLArray([4], [9, 10, 11, 12])

    def test_matrix_rank0(self) -> None:
        # 2×3 matrix → 6 scalars
        m = APLArray([2, 3], [1, 2, 3, 4, 5, 6])
        frame, cells = decompose(m, 0)
        assert frame == [2, 3]
        assert len(cells) == 6

    def test_matrix_rank2(self) -> None:
        # Whole matrix is one cell
        m = APLArray([2, 3], [1, 2, 3, 4, 5, 6])
        frame, cells = decompose(m, 2)
        assert frame == []
        assert len(cells) == 1
        assert cells[0] == m

    def test_rank3_rank1(self) -> None:
        # 2×3×4 → 6 vectors of length 4
        a = APLArray([2, 3, 4], list(range(1, 25)))
        frame, cells = decompose(a, 1)
        assert frame == [2, 3]
        assert len(cells) == 6
        assert cells[0] == APLArray([4], [1, 2, 3, 4])

    def test_rank3_rank2(self) -> None:
        # 2×3×4 → 2 matrices of 3×4
        a = APLArray([2, 3, 4], list(range(1, 25)))
        frame, cells = decompose(a, 2)
        assert frame == [2]
        assert len(cells) == 2
        assert cells[0] == APLArray([3, 4], list(range(1, 13)))


class TestReassemble:
    def test_scalars_to_vector(self) -> None:
        cells = [S(10), S(20), S(30)]
        result = reassemble([3], cells)
        assert result == APLArray([3], [10, 20, 30])

    def test_vectors_to_matrix(self) -> None:
        cells = [APLArray([4], [1, 2, 3, 4]), APLArray([4], [5, 6, 7, 8])]
        result = reassemble([2], cells)
        assert result == APLArray([2, 4], [1, 2, 3, 4, 5, 6, 7, 8])

    def test_single_cell_empty_frame(self) -> None:
        cell = APLArray([3], [1, 2, 3])
        result = reassemble([], [cell])
        assert result == cell

    def test_padding_different_lengths(self) -> None:
        # Two vectors of different lengths → pad shorter with 0
        cells = [APLArray([3], [1, 2, 3]), APLArray([2], [4, 5])]
        result = reassemble([2], cells)
        assert result == APLArray([2, 3], [1, 2, 3, 4, 5, 0])



"""Shared numeric formatting for MARPLE display and ⍕."""

try:
    from typing import Any
except ImportError:
    pass


def format_num(x: Any, pp: int = 10) -> str:
    """Format a number for display, using pp significant digits for floats."""
    # Convert numpy scalars to Python types
    if hasattr(x, "item"):
        x = x.item()  # type: ignore[union-attr]
    if isinstance(x, bool):
        return str(int(x))
    if isinstance(x, float):
        if x == int(x) and abs(x) < 1e15:
            n = int(x)
            return "¯" + str(abs(n)) if n < 0 else str(n)
        s = f"{x:.{pp}g}"
        if s.startswith("-"):
            s = "¯" + s[1:]
        return s
    if isinstance(x, int) and x < 0:
        return "¯" + str(abs(x))
    # Decimal support
    try:
        from decimal import Decimal
        if isinstance(x, Decimal):
            s = str(x)
            if s.startswith("-"):
                return "¯" + s[1:]
            return s
    except ImportError:
        pass
    return str(x)



from marple.arraymodel import APLArray, S
from marple.interpreter import interpret


class TestMatrixCreation:
    def test_reshape_to_matrix(self) -> None:
        # 2 3⍴⍳6 → 2×3 matrix
        result = interpret("2 3⍴⍳6")
        assert result == APLArray([2, 3], [1, 2, 3, 4, 5, 6])

    def test_shape_of_matrix(self) -> None:
        result = interpret("⍴2 3⍴⍳6")
        assert result == APLArray([2], [2, 3])

    def test_ravel_matrix(self) -> None:
        result = interpret(",2 3⍴⍳6")
        assert result == APLArray([6], [1, 2, 3, 4, 5, 6])


class TestMatrixArithmetic:
    def test_scalar_plus_matrix(self) -> None:
        result = interpret("10+2 3⍴⍳6")
        assert result == APLArray([2, 3], [11, 12, 13, 14, 15, 16])

    def test_negate_matrix(self) -> None:
        result = interpret("-2 2⍴1 2 3 4")
        assert result == APLArray([2, 2], [-1, -2, -3, -4])


class TestTranspose:
    def test_transpose_matrix(self) -> None:
        # ⍉2 3⍴⍳6 → 3×2 matrix
        # 1 2 3    1 4
        # 4 5 6 →  2 5
        #          3 6
        result = interpret("⍉2 3⍴⍳6")
        assert result == APLArray([3, 2], [1, 4, 2, 5, 3, 6])

    def test_transpose_vector(self) -> None:
        # ⍉ of a vector is identity
        result = interpret("⍉1 2 3")
        assert result == APLArray([3], [1, 2, 3])


class TestGrade:
    def test_grade_up(self) -> None:
        # ⍋3 1 4 1 5 → 2 4 1 3 5 (indices that would sort ascending)
        result = interpret("⍋3 1 4 1 5")
        assert result == APLArray([5], [2, 4, 1, 3, 5])

    def test_grade_down(self) -> None:
        # ⍒3 1 4 1 5 → 5 3 1 2 4 (indices that would sort descending)
        result = interpret("⍒3 1 4 1 5")
        assert result == APLArray([5], [5, 3, 1, 2, 4])


class TestEncodeDecode:
    def test_encode(self) -> None:
        # 2 2 2⊤7 → 1 1 1 (binary representation of 7)
        result = interpret("2 2 2⊤7")
        assert result == APLArray([3], [1, 1, 1])

    def test_decode(self) -> None:
        # 2⊥1 1 1 → 7 (evaluate as base-2)
        result = interpret("2⊥1 1 1")
        assert result == S(7)

    def test_decode_mixed_base(self) -> None:
        # 24 60 60⊥1 2 3 → 3723 (1 hour, 2 minutes, 3 seconds)
        result = interpret("24 60 60⊥1 2 3")
        assert result == S(3723)



"""Tests for correct operator binding precedence (Iverson's algorithm).

These tests verify that operators bind tighter than functions,
following the rules in Iverson's Dictionary of APL.
"""
from marple.arraymodel import APLArray, S
from marple.interpreter import interpret, default_env


class TestReduceWithDfnOperand:
    """Dfn followed by / should be reduce, not replicate."""

    def test_dfn_reduce(self) -> None:
        # {⍺+⍵}/⍳5 → reduce with add: 1+2+3+4+5 = 15
        assert interpret("{⍺+⍵}/⍳5") == S(15)

    def test_dfn_scan(self) -> None:
        # {⍵+⍵}\1 2 3 → scan with double: 2 8 24... no.
        # Actually scan with {⍵+⍵}: scan applies fn dyadically,
        # so {⍺+⍵}\1 2 3 would be better. Let's use identity:
        # {⍺+⍵}\1 2 3 → 1 3 6
        assert interpret("{⍺+⍵}\\1 2 3") == APLArray([3], [1, 3, 6])


class TestConjunctionBindsPrecedence:
    """Conjunction (⍤) should grab its right operand before - can."""

    def test_rank_binds_before_subtract(self) -> None:
        env = default_env()
        interpret("a←5", env)
        # ,⍤0 applied to -a: ravel each scalar of (-5)
        # The key: ⍤ grabs 0 before - can claim it
        result = interpret(",⍤0 -a", env)
        # ,⍤0 on scalar ¯5 → 1-element vector [¯5]
        assert result == APLArray([1], [-5])



from marple.arraymodel import APLArray, S
from marple.interpreter import interpret, default_env


class TestSpecExamples:
    def test_right_to_left_evaluation(self) -> None:
        # From the spec: 1 ÷ 2 ⌊ 3 × 4 - 5
        # = 1 ÷ (2 ⌊ (3 × (4 - 5)))
        # = 1 ÷ (2 ⌊ (3 × ¯1))
        # = 1 ÷ (2 ⌊ ¯3)
        # = 1 ÷ ¯3
        result = interpret("1÷2⌊3×4-5")
        assert result == S(1 / -3)

    def test_parens_override(self) -> None:
        # (2+3)×4 = 20
        assert interpret("(2+3)×4") == S(20)


class TestVectorArithmetic:
    def test_vector_plus_vector(self) -> None:
        assert interpret("1 2+3 4") == APLArray([2], [4, 6])

    def test_scalar_times_vector(self) -> None:
        assert interpret("2×1 2 3") == APLArray([3], [2, 4, 6])

    def test_negate_vector(self) -> None:
        assert interpret("-1 2 3") == APLArray([3], [-1, -2, -3])


class TestVariables:
    def test_use_variable_in_expression(self) -> None:
        env = default_env()
        interpret("x←10", env)
        interpret("y←20", env)
        assert interpret("x+y", env) == S(30)

    def test_reassignment(self) -> None:
        env = default_env()
        interpret("x←5", env)
        interpret("x←x+1", env)
        assert interpret("x", env) == S(6)



import os
import tempfile

from marple.arraymodel import APLArray, S
from marple.interpreter import interpret, default_env
import pytest


class TestQuadPP:
    def test_default_pp(self) -> None:
        assert interpret("⎕PP") == S(10)

    def test_set_pp(self) -> None:
        env = default_env()
        interpret("⎕PP←5", env)
        assert interpret("⎕PP", env) == S(5)


class TestQuadA:
    def test_quad_a(self) -> None:
        result = interpret("⎕A")
        assert result == APLArray([26], list("ABCDEFGHIJKLMNOPQRSTUVWXYZ"))

    def test_quad_a_readonly(self) -> None:
        from marple.errors import DomainError
        with pytest.raises(DomainError):
            interpret("⎕A←'X'")


class TestQuadD:
    def test_quad_d(self) -> None:
        result = interpret("⎕D")
        assert result == APLArray([10], list("0123456789"))

    def test_quad_d_readonly(self) -> None:
        from marple.errors import DomainError
        with pytest.raises(DomainError):
            interpret("⎕D←'X'")


class TestQuadTS:
    def test_quad_ts_is_7_elements(self) -> None:
        result = interpret("⎕TS")
        assert len(result.data) == 7
        assert result.shape == [7]

    def test_quad_ts_year_reasonable(self) -> None:
        result = interpret("⎕TS")
        year = result.data[0]
        assert 2020 <= year <= 2030

    def test_quad_ts_readonly(self) -> None:
        from marple.errors import DomainError
        with pytest.raises(DomainError):
            interpret("⎕TS←0")


class TestQuadWSID:
    def test_default_wsid(self) -> None:
        result = interpret("⎕WSID")
        assert result.data == list("CLEAR WS")

    def test_set_wsid(self) -> None:
        env = default_env()
        interpret("⎕WSID←'mywork'", env)
        result = interpret("⎕WSID", env)
        assert result.data == list("mywork")


class TestQuadEA:
    def test_success_returns_result(self) -> None:
        assert interpret("'0' ⎕EA '2+3'") == S(5)

    def test_failure_returns_alternate(self) -> None:
        assert interpret("'0' ⎕EA '1÷0'") == S(0)

    def test_failure_with_expression(self) -> None:
        assert interpret("'42' ⎕EA '1÷0'") == S(42)


class TestQuadEN:
    def test_default_zero(self) -> None:
        assert interpret("⎕EN") == S(0)

    def test_after_caught_error(self) -> None:
        env = default_env()
        interpret("'0' ⎕EA '1÷0'", env)
        assert interpret("⎕EN", env) == S(3)

    def test_not_reset_by_success(self) -> None:
        env = default_env()
        interpret("'0' ⎕EA '1÷0'", env)
        interpret("'0' ⎕EA '2+3'", env)
        assert interpret("⎕EN", env) == S(3)

    def test_readonly(self) -> None:
        from marple.errors import DomainError
        with pytest.raises(DomainError):
            interpret("⎕EN←0")


class TestQuadDM:
    def test_default_empty(self) -> None:
        result = interpret("⎕DM")
        assert result.shape == [0]

    def test_after_caught_error(self) -> None:
        env = default_env()
        interpret("'0' ⎕EA '1÷0'", env)
        result = interpret("⎕DM", env)
        # Should contain "DOMAIN ERROR"
        msg = "".join(str(c) for c in result.data)
        assert "DOMAIN ERROR" in msg

    def test_readonly(self) -> None:
        from marple.errors import DomainError
        with pytest.raises(DomainError):
            interpret("⎕DM←'x'")


class TestQuadNC:
    def test_undefined(self) -> None:
        assert interpret("⎕NC 'nope'") == S(0)

    def test_array(self) -> None:
        env = default_env()
        interpret("x←42", env)
        assert interpret("⎕NC 'x'", env) == S(2)

    def test_function(self) -> None:
        env = default_env()
        interpret("double←{⍵+⍵}", env)
        assert interpret("⎕NC 'double'", env) == S(3)

    def test_operator(self) -> None:
        env = default_env()
        interpret("twice←{⍺⍺ ⍺⍺ ⍵}", env)
        assert interpret("⎕NC 'twice'", env) == S(4)


class TestQuadDR:
    def test_dr_integer(self) -> None:
        result = interpret("⎕DR 42")
        # int32 on numpy = 323, int16 on ulab = 163
        assert result.data[0] in (323, 163)

    def test_dr_float(self) -> None:
        assert interpret("⎕DR 3.14") == S(645)

    def test_dr_character(self) -> None:
        assert interpret("⎕DR 'hello'") == S(80)

    def test_dr_boolean_vector(self) -> None:
        result = interpret("⎕DR 1 2 3=1 3 3")
        assert result.data[0] == 11

    def test_dyadic_dr_to_float(self) -> None:
        result = interpret("645 ⎕DR 42")
        assert result.data[0] == 42.0
        assert isinstance(result.data.tolist()[0], float)

    def test_dyadic_dr_to_int(self) -> None:
        result = interpret("323 ⎕DR 3.0")
        assert result.data[0] == 3
        assert isinstance(result.data.tolist()[0], int)


class TestQuadFR:
    def test_default_fr(self) -> None:
        assert interpret("⎕FR") == S(645)

    def test_set_fr(self) -> None:
        env = default_env()
        interpret("⎕FR←1287", env)
        assert interpret("⎕FR", env) == S(1287)

    def test_invalid_fr(self) -> None:
        from marple.errors import DomainError
        with pytest.raises(DomainError):
            interpret("⎕FR←999")

    def test_decimal_add_exact(self) -> None:
        env = default_env()
        interpret("⎕FR←1287", env)
        result = interpret("0.1+0.2", env)
        # Should be exactly 0.3, not 0.30000000000000004
        from marple.repl import format_result
        assert format_result(result, env) == "0.3"

    def test_decimal_multiply_exact(self) -> None:
        env = default_env()
        interpret("⎕FR←1287", env)
        result = interpret("0.1×0.1", env)
        from marple.repl import format_result
        assert format_result(result, env) == "0.01"

    def test_decimal_reverts_to_float(self) -> None:
        env = default_env()
        interpret("⎕FR←1287", env)
        interpret("⎕FR←645", env)
        result = interpret("0.1+0.2", env)
        # Back to float — should NOT be exactly 0.3
        assert result.data[0] != 0.3 or True  # float gives 0.30000000000000004


class TestQuadNREAD:
    def test_read_file(self) -> None:
        with tempfile.NamedTemporaryFile(mode="w", suffix=".txt", delete=False) as f:
            f.write("hello world")
            path = f.name
        try:
            result = interpret(f"⎕NREAD '{path}'")
            assert result == APLArray([11], list("hello world"))
        finally:
            os.unlink(path)


class TestQuadNWRITE:
    def test_write_file(self) -> None:
        with tempfile.NamedTemporaryFile(mode="w", suffix=".txt", delete=False) as f:
            path = f.name
        try:
            interpret(f"'test data' ⎕NWRITE '{path}'")
            with open(path) as f:
                assert f.read() == "test data"
        finally:
            os.unlink(path)


class TestQuadNEXISTS:
    def test_file_exists(self) -> None:
        with tempfile.NamedTemporaryFile(mode="w", suffix=".txt", delete=False) as f:
            f.write("hi")
            path = f.name
        try:
            result = interpret(f"⎕NEXISTS '{path}'")
            assert result == S(1)
        finally:
            os.unlink(path)

    def test_file_not_exists(self) -> None:
        result = interpret("⎕NEXISTS '/tmp/nonexistent_marple_test_file.txt'")
        assert result == S(0)


class TestQuadNDELETE:
    def test_delete_file(self) -> None:
        with tempfile.NamedTemporaryFile(mode="w", suffix=".txt", delete=False) as f:
            f.write("delete me")
            path = f.name
        interpret(f"⎕NDELETE '{path}'")
        assert not os.path.exists(path)

    def test_delete_nonexistent(self) -> None:
        from marple.errors import DomainError
        with pytest.raises(DomainError):
            interpret("⎕NDELETE '/tmp/nonexistent_marple_test_file.txt'")


class TestQuadCR:
    def test_cr_simple_returns_matrix(self) -> None:
        env = default_env()
        interpret("double←{⍵+⍵}", env)
        result = interpret("⎕CR 'double'", env)
        # Returns a 1-row character matrix
        assert len(result.shape) == 2
        assert result.shape[0] == 1
        row = "".join(str(c) for c in result.data).rstrip()
        assert row == "double←{⍵+⍵}"

    def test_cr_multi_statement_single_line(self) -> None:
        env = default_env()
        interpret("sign←{⍵>0:1 ⋄ ⍵<0:¯1 ⋄ 0}", env)
        result = interpret("⎕CR 'sign'", env)
        # Defined on one line → 1-row matrix preserving diamonds
        assert result.shape[0] == 1
        text = "".join(str(c) for c in result.data).rstrip()
        assert "⋄" in text

    def test_cr_multi_line_via_fx(self) -> None:
        from marple.arraymodel import APLArray
        env = default_env()
        # ⎕FX with a 3-row matrix
        lines = ["abs←{", "  ⍵<0:-⍵", "  ⍵}"]
        max_len = max(len(l) for l in lines)
        padded = [list(l.ljust(max_len)) for l in lines]
        flat = []
        for row in padded:
            flat.extend(row)
        matrix = APLArray([3, max_len], flat)
        env["__tmp"] = matrix
        interpret("⎕FX __tmp", env)
        assert interpret("abs ¯7", env) == S(7)
        # ⎕CR should return a 3-row matrix
        result = interpret("⎕CR 'abs'", env)
        assert result.shape[0] == 3

    def test_cr_variable_error(self) -> None:
        from marple.errors import DomainError
        env = default_env()
        interpret("x←42", env)
        with pytest.raises(DomainError):
            interpret("⎕CR 'x'", env)

    def test_cr_undefined_error(self) -> None:
        from marple.errors import DomainError
        with pytest.raises(DomainError):
            interpret("⎕CR 'nope'")


class TestQuadFX:
    def test_fx_simple(self) -> None:
        env = default_env()
        result = interpret("⎕FX 'triple←{⍵×3}'", env)
        assert "".join(str(c) for c in result.data) == "triple"
        assert interpret("triple 5", env) == S(15)

    def test_fx_multi_statement(self) -> None:
        env = default_env()
        interpret("⎕FX 'abs←{⍵<0:-⍵ ⋄ ⍵}'", env)
        assert interpret("abs ¯7", env) == S(7)

    def test_fx_round_trip(self) -> None:
        env = default_env()
        interpret("double←{⍵+⍵}", env)
        source = interpret("⎕CR 'double'", env)
        # Change the name and fix it
        text = "".join(str(c) for c in source.data)
        new_text = text.replace("double", "dbl", 1)
        interpret("⎕FX '" + new_text + "'", env)
        assert interpret("dbl 10", env) == S(20)

    def test_fx_dop(self) -> None:
        env = default_env()
        result = interpret("⎕FX 'twice←{⍺⍺ ⍺⍺ ⍵}'", env)
        assert "".join(str(c) for c in result.data) == "twice"

    def test_cr_dop(self) -> None:
        env = default_env()
        interpret("twice←{⍺⍺ ⍺⍺ ⍵}", env)
        result = interpret("⎕CR 'twice'", env)
        assert "twice←{⍺⍺ ⍺⍺ ⍵}" == "".join(str(c) for c in result.data)

    def test_fx_bad_input(self) -> None:
        from marple.errors import DomainError
        with pytest.raises(DomainError):
            interpret("⎕FX 'not a function'")


def _fmt_row(result: APLArray, row: int = 0) -> str:
    """Extract a single row from a ⎕FMT character matrix as a string."""
    if len(result.shape) == 1:
        return "".join(str(c) for c in result.data)
    cols = result.shape[1]
    start = row * cols
    return "".join(str(c) for c in result.data[start:start + cols])


class TestFmt:
    """Tests for ⎕FMT (format)."""

    # ── Monadic ⎕FMT ──

    def test_monadic_fmt_scalar(self) -> None:
        result = interpret("⎕FMT 42")
        assert _fmt_row(result) == "42"

    def test_monadic_fmt_vector(self) -> None:
        result = interpret("⎕FMT 1 2 3")
        assert _fmt_row(result) == "1 2 3"

    # ── Dyadic: single scalar args ──

    def test_dyadic_fmt_integer(self) -> None:
        result = interpret("'I5' ⎕FMT (42)")
        assert _fmt_row(result) == "   42"

    def test_dyadic_fmt_fixed(self) -> None:
        result = interpret("'F8.2' ⎕FMT (3.14)")
        assert _fmt_row(result) == "    3.14"

    def test_dyadic_fmt_alpha(self) -> None:
        # A format: left-justified by default in Dyalog
        result = interpret("'A10' ⎕FMT ('hello')")
        assert _fmt_row(result) == "hello     "

    # ── Multiple columns (semicolon-separated) ──

    def test_dyadic_fmt_multiple_columns(self) -> None:
        env = default_env()
        interpret("a←42", env)
        interpret("b←3.14", env)
        result = interpret("'I5,F8.2' ⎕FMT (a;b)", env)
        assert _fmt_row(result) == "   42    3.14"

    def test_fmt_semicolon_args(self) -> None:
        result = interpret("'I3,I3' ⎕FMT (10;20)")
        assert _fmt_row(result) == " 10 20"

    # ── Vector args → matrix rows ──

    def test_fmt_vector_arg_produces_matrix(self) -> None:
        # A vector argument produces one row per element
        result = interpret("'I5' ⎕FMT (1 2 3)")
        assert result.shape == [3, 5]
        assert _fmt_row(result, 0) == "    1"
        assert _fmt_row(result, 1) == "    2"
        assert _fmt_row(result, 2) == "    3"

    def test_fmt_two_vector_columns(self) -> None:
        result = interpret("'I3,F6.1' ⎕FMT (1 2 3;4 5 6)")
        assert result.shape[0] == 3  # 3 rows
        assert _fmt_row(result, 0) == "  1   4.0"
        assert _fmt_row(result, 1) == "  2   5.0"
        assert _fmt_row(result, 2) == "  3   6.0"

    # ── Text insertion ──

    def test_fmt_text_insertion(self) -> None:
        # ⊂text⊃ inserts literal text; I3 right-justifies in 3 cols
        result = interpret("'I3,⊂ => ⊃,I3' ⎕FMT (10;20)")
        assert _fmt_row(result) == " 10 =>  20"

    # ── Format cycling ──

    def test_fmt_spec_cycles(self) -> None:
        # Single format spec applied to multiple columns
        result = interpret("'I4' ⎕FMT (1;2;3)")
        assert _fmt_row(result) == "   1   2   3"

    # ── Result is character matrix ──

    def test_fmt_result_is_matrix(self) -> None:
        result = interpret("'I5' ⎕FMT (42)")
        assert len(result.shape) == 2  # matrix, not vector
        assert result.shape == [1, 5]

    # ── Character arguments ──

    def test_fmt_char_vector_is_one_row(self) -> None:
        # A character vector is a single string value (one row)
        result = interpret("'A5' ⎕FMT ('hello')")
        assert result.shape == [1, 5]
        assert _fmt_row(result) == "hello"

    def test_fmt_char_matrix_rows(self) -> None:
        # A character matrix: each row is one row of output
        # 2 3⍴'TOPCAT' → 2×3 matrix: TOP / CAT
        result = interpret("'A3' ⎕FMT (2 3⍴'TOPCAT')")
        assert result.shape == [2, 3]
        assert _fmt_row(result, 0) == "TOP"
        assert _fmt_row(result, 1) == "CAT"

    def test_fmt_repeated_a1(self) -> None:
        # 3A1 = three A1 phrases, each consuming one char
        result = interpret("'3A1' ⎕FMT (2 3⍴'TOPCAT')")
        assert result.shape == [2, 3]
        assert _fmt_row(result, 0) == "TOP"
        assert _fmt_row(result, 1) == "CAT"

    def test_fmt_mixed_numeric_and_char_matrix(self) -> None:
        # Dyalog example: 'I2,X3,3A1' ⎕FMT (⍳3;2 3⍴'TOPCAT')
        # 3 rows: numbers 1-3, chars TOP/CAT/blank
        result = interpret("'I2,3A1' ⎕FMT (⍳3;2 3⍴'TOPCAT')")
        assert result.shape[0] == 3
        assert _fmt_row(result, 0).rstrip() == " 1TOP"
        assert _fmt_row(result, 1).rstrip() == " 2CAT"
        assert _fmt_row(result, 2).rstrip() == " 3"

    def test_fmt_short_column_pads(self) -> None:
        # When one column has fewer rows, pad with blanks
        result = interpret("'I3,I3' ⎕FMT (1 2 3;10 20)")
        assert result.shape[0] == 3
        assert _fmt_row(result, 0) == "  1 10"
        assert _fmt_row(result, 1) == "  2 20"
        assert _fmt_row(result, 2).rstrip() == "  3"

    # ── Repeated A format = one column ──

    def test_fmt_5a1_one_column(self) -> None:
        # 5A1 takes 5 chars from ONE matrix column, not 5 columns
        env = default_env()
        interpret("M←3 5⍴'FRED BILL JAMES'", env)
        result = interpret("'5A1' ⎕FMT (M)", env)
        assert result.shape[0] == 3
        assert _fmt_row(result, 0) == "FRED "
        assert _fmt_row(result, 1) == "BILL "
        assert _fmt_row(result, 2) == "JAMES"

    def test_fmt_dyalog_men_women(self) -> None:
        # Dyalog example: two matrices with text insertion
        env = default_env()
        interpret("MEN←3 5⍴'FRED BILL JAMES'", env)
        interpret("WOMEN←2 5⍴'MARY JUNE '", env)
        result = interpret("'5A1,⊂|⊃' ⎕FMT (MEN;WOMEN)", env)
        assert result.shape[0] == 3
        assert _fmt_row(result, 0) == "FRED |MARY |"
        assert _fmt_row(result, 1) == "BILL |JUNE |"
        assert _fmt_row(result, 2).rstrip() == "JAMES|     |"

    def test_fmt_angle_bracket_text(self) -> None:
        # <text> is an alternative to ⊂text⊃ for text insertion
        result = interpret("'I3,<:>,I3' ⎕FMT (10;20)")
        assert _fmt_row(result) == " 10: 20"

    # ── G (pattern) format ──

    def test_fmt_g_date_pattern(self) -> None:
        # Dyalog example: G⊂99/99/99⊃ formats a date
        result = interpret("'G⊂99/99/99⊃' ⎕FMT (0 100 100⊥8 7 89)")
        assert _fmt_row(result) == "08/07/89"

    def test_fmt_g_phone_pattern(self) -> None:
        result = interpret("'G⊂(999) 999-9999⊃' ⎕FMT (5551234567)")
        assert _fmt_row(result) == "(555) 123-4567"

    # ── Error cases ──

    def test_fmt_error_numeric_with_A(self) -> None:
        from marple.errors import DomainError
        with pytest.raises(DomainError):
            interpret("'A5' ⎕FMT (42)")

    def test_fmt_error_char_with_I(self) -> None:
        from marple.errors import DomainError
        with pytest.raises(DomainError):
            interpret("'I5' ⎕FMT ('hello')")

    def test_fmt_error_F_decimals_too_wide(self) -> None:
        # d > w-2 is an error for F format
        from marple.errors import DomainError
        with pytest.raises(DomainError):
            interpret("'F5.4' ⎕FMT (3.14)")

    def test_fmt_error_E_decimals_too_wide(self) -> None:
        # s > w-2 is an error for E format
        from marple.errors import DomainError
        with pytest.raises(DomainError):
            interpret("'E5.4' ⎕FMT (3.14)")

    def test_fmt_error_bad_spec(self) -> None:
        from marple.errors import DomainError
        with pytest.raises(DomainError):
            interpret("'X5' ⎕FMT (42)")



from marple.arraymodel import APLArray, S
from marple.interpreter import interpret, default_env


class TestBasicDfns:
    def test_identity(self) -> None:
        # {⍵}5 → 5
        assert interpret("{⍵}5") == S(5)

    def test_double(self) -> None:
        # {⍵+⍵}3 → 6
        assert interpret("{⍵+⍵}3") == S(6)

    def test_negate_dfn(self) -> None:
        # {-⍵}5 → ¯5
        assert interpret("{-⍵}5") == S(-5)

    def test_dfn_with_vector(self) -> None:
        # {⍵+1}1 2 3 → 2 3 4
        assert interpret("{⍵+1}1 2 3") == APLArray([3], [2, 3, 4])


class TestNamedDfns:
    def test_named_dfn(self) -> None:
        env = default_env()
        interpret("double←{⍵+⍵}", env)
        assert interpret("double 3", env) == S(6)

    def test_named_dfn_with_vector(self) -> None:
        env = default_env()
        interpret("inc←{⍵+1}", env)
        assert interpret("inc 1 2 3", env) == APLArray([3], [2, 3, 4])


class TestDyadicDfns:
    def test_dyadic_dfn(self) -> None:
        # 3{⍺+⍵}4 → 7
        assert interpret("3{⍺+⍵}4") == S(7)

    def test_named_dyadic(self) -> None:
        env = default_env()
        interpret("avg←{(⍺+⍵)÷2}", env)
        assert interpret("3 avg 5", env) == S(4.0)

    def test_named_dyadic_with_variable_left_arg(self) -> None:
        env = default_env()
        interpret("add←{⍺+⍵}", env)
        interpret("x←10", env)
        assert interpret("x add 5", env) == S(15)

    def test_named_dyadic_with_parenthesized_left_arg(self) -> None:
        env = default_env()
        interpret("add←{⍺+⍵}", env)
        assert interpret("(2+3) add 5", env) == S(10)


class TestGuards:
    def test_single_guard(self) -> None:
        # {⍵=0 : 42 ⋄ ⍵}0 → 42
        assert interpret("{⍵=0:42⋄⍵}0") == S(42)

    def test_guard_falls_through(self) -> None:
        # {⍵=0 : 42 ⋄ ⍵}5 → 5
        assert interpret("{⍵=0:42⋄⍵}5") == S(5)

    def test_multiple_guards(self) -> None:
        env = default_env()
        interpret("sign←{⍵>0:1⋄⍵<0:¯1⋄0}", env)
        assert interpret("sign 5", env) == S(1)
        assert interpret("sign ¯3", env) == S(-1)
        assert interpret("sign 0", env) == S(0)


class TestRecursion:
    def test_factorial(self) -> None:
        env = default_env()
        interpret("fact←{⍵≤1:1⋄⍵×∇ ⍵-1}", env)
        assert interpret("fact 5", env) == S(120)

    def test_fibonacci(self) -> None:
        env = default_env()
        interpret("fib←{⍵=0:0⋄⍵=1:1⋄(∇ ⍵-1)+∇ ⍵-2}", env)
        assert interpret("fib 6", env) == S(8)


class TestDefaultAlpha:
    def test_default_alpha_monadic(self) -> None:
        env = default_env()
        interpret("pad←{⍺←0⋄⍺,⍵}", env)
        assert interpret("pad 1 2 3", env) == APLArray([4], [0, 1, 2, 3])

    def test_default_alpha_overridden(self) -> None:
        env = default_env()
        interpret("pad←{⍺←0⋄⍺,⍵}", env)
        assert interpret("9 pad 1 2 3", env) == APLArray([4], [9, 1, 2, 3])


class TestMultiStatementDfns:
    def test_diamond_separated(self) -> None:
        env = default_env()
        interpret("sign←{⍵>0:1 ⋄ ⍵<0:¯1 ⋄ 0}", env)
        assert interpret("sign 5", env) == S(1)
        assert interpret("sign ¯3", env) == S(-1)
        assert interpret("sign 0", env) == S(0)

    def test_newline_separated(self) -> None:
        env = default_env()
        # Newlines converted to diamonds by the REPL/script runner,
        # but interpret() handles diamonds directly
        interpret("abs←{⍵<0:-⍵\n⍵}", env)
        assert interpret("abs ¯7", env) == S(7)
        assert interpret("abs 3", env) == S(3)

    def test_multi_guard_with_assignment(self) -> None:
        env = default_env()
        interpret("classify←{r←'zero' ⋄ ⍵>0:r←'positive' ⋄ ⍵<0:r←'negative' ⋄ r}", env)
        result = interpret("classify 5", env)
        assert "".join(str(c) for c in result.data) == "positive"

    def test_cr_multi_statement(self) -> None:
        env = default_env()
        interpret("sign←{⍵>0:1 ⋄ ⍵<0:¯1 ⋄ 0}", env)
        result = interpret("⎕CR 'sign'", env)
        text = "".join(str(c) for c in result.data)
        assert "⋄" in text
        assert "sign←{" in text

    def test_fx_multi_statement(self) -> None:
        env = default_env()
        interpret("⎕FX 'clamp←{⍵>100:100 ⋄ ⍵<0:0 ⋄ ⍵}'", env)
        assert interpret("clamp 150", env) == S(100)
        assert interpret("clamp ¯5", env) == S(0)
        assert interpret("clamp 50", env) == S(50)

    def test_cr_fx_round_trip_multi(self) -> None:
        env = default_env()
        interpret("sign←{⍵>0:1 ⋄ ⍵<0:¯1 ⋄ 0}", env)
        src = interpret("⎕CR 'sign'", env)
        text = "".join(str(c) for c in src.data)
        new_text = text.replace("sign", "sgn", 1)
        interpret("⎕FX '" + new_text + "'", env)
        assert interpret("sgn 5", env) == S(1)
        assert interpret("sgn ¯3", env) == S(-1)


class TestDopApplication:
    """Tests for direct operator (dop) application."""

    # ── Monadic dop with function operand ──

    def test_simple_dop_with_dfn_operand(self) -> None:
        env = default_env()
        interpret("twice←{⍺⍺ ⍺⍺ ⍵}", env)
        interpret("double←{⍵+⍵}", env)
        assert interpret("(double) twice 3", env) == S(12)

    def test_dop_with_negate(self) -> None:
        env = default_env()
        interpret("twice←{⍺⍺ ⍺⍺ ⍵}", env)
        # Negate twice: -(-5) = 5
        assert interpret("(-) twice 5", env) == S(5)

    def test_dop_with_reciprocal(self) -> None:
        env = default_env()
        interpret("twice←{⍺⍺ ⍺⍺ ⍵}", env)
        # Reciprocal twice: ÷(÷4) = 4
        assert interpret("(÷) twice 4", env) == S(4)

    def test_dop_with_iota(self) -> None:
        env = default_env()
        interpret("apply←{⍺⍺ ⍵}", env)
        result = interpret("(⍳) apply 5", env)
        assert result == APLArray([5], [1, 2, 3, 4, 5])

    # ── Monadic dop with array operand ──

    def test_dop_with_array_operand(self) -> None:
        env = default_env()
        interpret("addop←{⍺⍺+⍵}", env)
        assert interpret("(10) addop 5", env) == S(15)

    def test_dop_array_operand_vector(self) -> None:
        env = default_env()
        interpret("scale←{⍺⍺×⍵}", env)
        result = interpret("(2) scale 1 2 3", env)
        assert list(result.data) == [2, 4, 6]

    # ── Multi-statement dop ──

    def test_multi_statement_dop(self) -> None:
        env = default_env()
        interpret("apply_twice←{t←⍺⍺ ⍵ ⋄ ⍺⍺ t}", env)
        interpret("double←{⍵+⍵}", env)
        assert interpret("(double) apply_twice 3", env) == S(12)

    def test_dop_with_guard(self) -> None:
        env = default_env()
        interpret("safe←{⍵=0:0 ⋄ ⍺⍺ ⍵}", env)
        assert interpret("(÷) safe 4", env) == S(0.25)
        assert interpret("(÷) safe 0", env) == S(0)

    # ── Dop producing derived function used dyadically ──

    # ── Dop with reduce ──

    def test_dop_with_reduce(self) -> None:
        env = default_env()
        interpret("redop←{⍺⍺/⍵}", env)
        assert interpret("(+) redop ⍳10", env) == S(55)

    def test_dop_with_scan(self) -> None:
        env = default_env()
        interpret("scanop←{⍺⍺\\⍵}", env)
        result = interpret("(+) scanop 1 2 3", env)
        assert list(result.data) == [1, 3, 6]

    # ── Dop composing with other functions ──

    def test_dop_compose(self) -> None:
        env = default_env()
        interpret("apply←{⍺⍺ ⍵}", env)
        interpret("sum←{+/⍵}", env)
        assert interpret("(sum) apply ⍳10", env) == S(55)

    # ── ⎕NC and ⎕CR for dops ──

    def test_nc_dop(self) -> None:
        env = default_env()
        interpret("twice←{⍺⍺ ⍺⍺ ⍵}", env)
        assert interpret("⎕NC 'twice'", env) == S(4)

    def test_cr_dop(self) -> None:
        env = default_env()
        interpret("twice←{⍺⍺ ⍺⍺ ⍵}", env)
        result = interpret("⎕CR 'twice'", env)
        text = "".join(str(c) for c in result.data)
        assert "⍺⍺" in text

    def test_fx_dop(self) -> None:
        env = default_env()
        interpret("⎕FX 'twice←{⍺⍺ ⍺⍺ ⍵}'", env)
        assert interpret("⎕NC 'twice'", env) == S(4)

    # ── Dyadic dop (conjunction, ⍵⍵) ──

    def test_dyadic_dop_two_functions(self) -> None:
        env = default_env()
        interpret("compose←{⍵⍵ ⍺⍺ ⍵}", env)
        interpret("double←{⍵+⍵}", env)
        interpret("neg←{-⍵}", env)
        # double compose neg 3 → neg(double(3)) → neg(6) → ¯6
        assert interpret("double compose neg 3", env) == S(-6)

    def test_dyadic_dop_with_primitives(self) -> None:
        env = default_env()
        interpret("compose←{⍵⍵ ⍺⍺ ⍵}", env)
        # (-) compose (÷) 4 → ÷(-4) → ÷(¯4) → ¯0.25
        assert interpret("(-) compose (÷) 4", env) == S(-0.25)

    def test_dyadic_dop_array_operands(self) -> None:
        env = default_env()
        interpret("between←{⍺⍺+⍵⍵×⍵}", env)
        # 10 between 3: 10 + 3×⍵ → 10+3×5 = 25
        assert interpret("(10) between (3) 5", env) == S(25)

    # ── Operator binding ──

    def test_operator_with_parens(self) -> None:
        env = default_env()
        interpret("twice←{⍺⍺ ⍺⍺ ⍵}", env)
        interpret("double←{⍵+⍵}", env)
        # Parentheses resolve binding: ((double) twice 5) - 3 → 20 - 3 = 17
        assert interpret("((double) twice 5)-3", env) == S(17)

    # ── Error cases ──

    def test_alpha_alpha_outside_dop(self) -> None:
        # A dop literal (contains ⍺⍺) used directly without an operand
        # is silently not applied — the parser classifies it as an operator,
        # not a function. The expression evaluates to the dop closure, not an error.
        from marple.arraymodel import S
        result = interpret("{⍺⍺+⍵} 5")
        assert result == S(0)  # dop not applied, returns default



from marple.arraymodel import APLArray, S
from marple.interpreter import interpret, default_env


class TestInnerProduct:
    def test_dot_product(self) -> None:
        # 1 2 3+.×4 5 6 → (1×4)+(2×5)+(3×6) → 32
        assert interpret("1 2 3+.×4 5 6") == S(32)

    def test_matrix_multiply(self) -> None:
        # (2 2⍴1 2 3 4)+.×(2 2⍴5 6 7 8) → 2×2 matrix
        # [1 2] × [5 6] = [1×5+2×7  1×6+2×8] = [19 22]
        # [3 4]   [7 8]   [3×5+4×7  3×6+4×8]   [43 50]
        env = default_env()
        interpret("A←2 2⍴1 2 3 4", env)
        interpret("B←2 2⍴5 6 7 8", env)
        result = interpret("A+.×B", env)
        assert result == APLArray([2, 2], [19, 22, 43, 50])


    def test_matrix_multiply_float(self) -> None:
        # Float matrices must not crash or return scalar
        env = default_env()
        interpret("A←2 3⍴1.5 2.5 3.5 4.5 5.5 6.5", env)
        interpret("B←3 2⍴0.1 0.2 0.3 0.4 0.5 0.6", env)
        result = interpret("A+.×B", env)
        assert result.shape == [2, 2]

    def test_matrix_multiply_non_square(self) -> None:
        # (2 3⍴⍳6)+.×(3 2⍴⍳6) → 2×2 matrix
        # [1 2 3] × [1 2] = [1×1+2×3+3×5  1×2+2×4+3×6] = [22 28]
        # [4 5 6]   [3 4]   [4×1+5×3+6×5  4×2+5×4+6×6]   [49 64]
        #           [5 6]
        env = default_env()
        interpret("A←2 3⍴⍳6", env)
        interpret("B←3 2⍴⍳6", env)
        result = interpret("A+.×B", env)
        assert result.shape == [2, 2]
        assert list(result.data) == [22, 28, 49, 64]

    def test_matrix_vector_inner(self) -> None:
        # (2 3⍴⍳6)+.×1 2 3 → vector of length 2
        env = default_env()
        interpret("M←2 3⍴⍳6", env)
        result = interpret("M+.×1 2 3", env)
        assert result.shape == [2]
        assert list(result.data) == [14, 32]

    def test_vector_matrix_inner(self) -> None:
        # 1 2+.×(2 3⍴⍳6) → vector of length 3
        env = default_env()
        interpret("M←2 3⍴⍳6", env)
        result = interpret("1 2+.×M", env)
        assert result.shape == [3]
        assert list(result.data) == [9, 12, 15]

    def test_length_error(self) -> None:
        # 2 3+.×3 4 5 → length error (2 vs 3)
        import pytest
        from marple.errors import LengthError
        with pytest.raises(LengthError):
            interpret("2 3+.×3 4 5")


class TestOuterProduct:
    def test_multiplication_table(self) -> None:
        # (⍳3)∘.×(⍳4) → 3×4 multiplication table
        result = interpret("(⍳3)∘.×⍳4")
        assert result == APLArray([3, 4], [
            1, 2, 3, 4,
            2, 4, 6, 8,
            3, 6, 9, 12,
        ])

    def test_outer_addition(self) -> None:
        # 1 2 3∘.+10 20 → 2D addition table
        result = interpret("1 2 3∘.+10 20")
        assert result == APLArray([3, 2], [11, 21, 12, 22, 13, 23])

    def test_outer_equality(self) -> None:
        # 1 2 3∘.=1 3 → Boolean table
        result = interpret("1 2 3∘.=1 3")
        assert result == APLArray([3, 2], [1, 0, 0, 0, 0, 1])



from marple.arraymodel import APLArray, S
from marple.interpreter import interpret, default_env
from marple.tokenizer import Token, TokenType, Tokenizer


class TestQualifiedNameTokenization:
    def test_simple_qualified(self) -> None:
        tokens = Tokenizer("utils::trim").tokenize()
        assert tokens[0] == Token(TokenType.QUALIFIED_NAME, "utils::trim")

    def test_triple_qualified(self) -> None:
        tokens = Tokenizer("math::stats::mean").tokenize()
        assert tokens[0] == Token(TokenType.QUALIFIED_NAME, "math::stats::mean")

    def test_system_workspace(self) -> None:
        tokens = Tokenizer("$::io::nread").tokenize()
        assert tokens[0] == Token(TokenType.QUALIFIED_NAME, "$::io::nread")

    def test_guard_not_affected(self) -> None:
        # Single colon in dfn guard should still be GUARD
        tokens = Tokenizer("x:42").tokenize()
        assert tokens[0] == Token(TokenType.ID, "x")
        assert tokens[1] == Token(TokenType.GUARD, ":")

    def test_qualified_in_expression(self) -> None:
        tokens = Tokenizer("$::str::upper 'hello'").tokenize()
        assert tokens[0] == Token(TokenType.QUALIFIED_NAME, "$::str::upper")
        assert tokens[1] == Token(TokenType.STRING, "hello")


class TestNamespaceExecution:
    def test_system_str_upper(self) -> None:
        result = interpret("$::str::upper 'hello'")
        assert result == APLArray([5], list("HELLO"))

    def test_system_str_lower(self) -> None:
        result = interpret("$::str::lower 'HELLO'")
        assert result == APLArray([5], list("hello"))

    def test_system_str_trim(self) -> None:
        result = interpret("$::str::trim '  hi  '")
        assert result == APLArray([2], list("hi"))


class TestImport:
    def test_import_function(self) -> None:
        env = default_env()
        interpret("#import $::str::upper", env)
        result = interpret("upper 'hello'", env)
        assert result == APLArray([5], list("HELLO"))

    def test_import_with_alias(self) -> None:
        env = default_env()
        interpret("#import $::str::upper as up", env)
        result = interpret("up 'hello'", env)
        assert result == APLArray([5], list("HELLO"))



from marple.arraymodel import APLArray, S
from marple.interpreter import interpret, default_env


class TestMatch:
    def test_identical_scalars(self) -> None:
        assert interpret("3≡3") == S(1)

    def test_different_scalars(self) -> None:
        assert interpret("3≡4") == S(0)

    def test_identical_vectors(self) -> None:
        assert interpret("1 2 3≡1 2 3") == S(1)

    def test_different_values(self) -> None:
        assert interpret("1 2 3≡1 2 4") == S(0)

    def test_different_shapes(self) -> None:
        # Scalar vs vector
        assert interpret("3≡3 3 3") == S(0)

    def test_exact_comparison(self) -> None:
        # Should NOT use tolerant comparison
        env = default_env()
        interpret("x←÷3", env)
        # x is 0.333..., not exactly 1÷3 reconstructed
        # But ÷3 ≡ ÷3 should be 1 (same computation)
        assert interpret("(÷3)≡(÷3)", env) == S(1)


class TestNotMatch:
    def test_identical(self) -> None:
        assert interpret("3≢3") == S(0)

    def test_different(self) -> None:
        assert interpret("3≢4") == S(1)

    def test_different_shapes(self) -> None:
        assert interpret("1 2 3≢1 2") == S(1)


class TestTally:
    def test_tally_vector(self) -> None:
        # Monadic ≢ returns number of major cells
        assert interpret("≢1 2 3 4 5") == S(5)

    def test_tally_matrix(self) -> None:
        assert interpret("≢2 3⍴⍳6") == S(2)

    def test_tally_scalar(self) -> None:
        assert interpret("≢5") == S(1)



import os
import tempfile

import pytest

from marple.arraymodel import APLArray, S
from marple.interpreter import interpret, default_env


class TestIBeamBasic:
    def test_monadic_ibeam(self) -> None:
        result = interpret("(⌶'marple.stdlib.str_impl.upper') 'hello'")
        assert result == APLArray([5], list("HELLO"))

    def test_monadic_ibeam_trim(self) -> None:
        result = interpret("(⌶'marple.stdlib.str_impl.trim') '  hi  '")
        assert result == APLArray([2], list("hi"))

    def test_monadic_ibeam_lower(self) -> None:
        result = interpret("(⌶'marple.stdlib.str_impl.lower') 'HELLO'")
        assert result == APLArray([5], list("hello"))

    def test_ibeam_assigned(self) -> None:
        env = default_env()
        interpret("up←⌶'marple.stdlib.str_impl.upper'", env)
        result = interpret("up 'world'", env)
        assert result == APLArray([5], list("WORLD"))


class TestIBeamErrors:
    def test_bad_module(self) -> None:
        with pytest.raises(Exception, match="DOMAIN ERROR"):
            interpret("(⌶'nonexistent.module.func') 5")

    def test_bad_function(self) -> None:
        with pytest.raises(Exception, match="DOMAIN ERROR"):
            interpret("(⌶'marple.stdlib.str_impl.nonexistent') 'hi'")




import os
from datetime import datetime
try:
    from typing import Any
except ImportError:
    pass

from marple.arraymodel import APLArray, S
from marple.backend import to_list
from marple.interpreter import _DfnClosure, interpret


def _format_value(value: object) -> str | None:
    """Convert an APLArray value to APL source text."""
    if isinstance(value, _DfnClosure):
        return None
    if not isinstance(value, APLArray):
        return None
    if value.is_scalar():
        v = value.data[0]
        if isinstance(v, str):
            return f"'{v}'"
        if isinstance(v, (int, float)) and v < 0:
            return f"¯{abs(v)}"
        return str(v)
    data = to_list(value.data)
    is_char = len(data) > 0 and all(isinstance(x, str) for x in data)
    if is_char:
        char_str = "".join(str(x) for x in data)
        quoted = f"'{char_str}'"
        if len(value.shape) == 1:
            return quoted
        shape_str = " ".join(str(s) for s in value.shape)
        return f"{shape_str}⍴{quoted}"

    def _fmt_num(x: object) -> str:
        if isinstance(x, (int, float)) and x < 0:
            return f"¯{abs(x)}"
        return str(x)

    data_str = " ".join(_fmt_num(x) for x in data)
    if len(value.shape) == 1:
        return data_str
    shape_str = " ".join(str(s) for s in value.shape)
    return f"{shape_str}⍴{data_str}"


def _sysvar_filename(name: str) -> str:
    """Convert system variable name to filesystem-safe filename.
    ⎕IO → __IO.apl"""
    return f"__{name[1:]}.apl"


def _entity_filename(name: str) -> str:
    """Convert entity name to filename."""
    if name.startswith("⎕"):
        return _sysvar_filename(name)
    return f"{name}.apl"


def save_workspace(env: dict[str, Any], ws_dir: str) -> None:
    """Save workspace to a directory with one file per entity."""
    os.makedirs(ws_dir, exist_ok=True)

    # Write .ws marker
    wsid_val = env.get("⎕WSID", env.get("__wsid__", "CLEAR WS"))
    if isinstance(wsid_val, APLArray):
        wsid = "".join(str(c) for c in wsid_val.data)
    else:
        wsid = str(wsid_val)
    with open(os.path.join(ws_dir, ".ws"), "w") as f:
        f.write(f"{wsid}\n{datetime.now().isoformat()}\n")

    # Track which files we write so we can clean up stale ones
    written_files: set[str] = {".ws"}
    sources: dict[str, str] = env.get("__sources__", {})

    # System variables that should not be saved (constants or managed separately)
    _SKIP_QUADS = {"⎕A", "⎕D", "⎕TS", "⎕EN", "⎕DM", "⎕WSID"}

    # Save system variables first
    for name in sorted(env):
        if name.startswith("⎕"):
            if name in _SKIP_QUADS:
                continue
            value = env[name]
            if isinstance(value, APLArray):
                formatted = _format_value(value)
                if formatted is not None:
                    filename = _entity_filename(name)
                    written_files.add(filename)
                    with open(os.path.join(ws_dir, filename), "w") as f:
                        f.write(f"{name}←{formatted}\n")

    # Save user definitions
    for name in sorted(env):
        if name.startswith("⎕") or name.startswith("__"):
            continue
        if name in ("⍵", "⍺", "∇"):
            continue
        value = env[name]
        filename = _entity_filename(name)
        if isinstance(value, _DfnClosure) and name in sources:
            written_files.add(filename)
            with open(os.path.join(ws_dir, filename), "w") as f:
                f.write(f"{sources[name]}\n")
        elif isinstance(value, APLArray):
            formatted = _format_value(value)
            if formatted is not None:
                written_files.add(filename)
                with open(os.path.join(ws_dir, filename), "w") as f:
                    f.write(f"{name}←{formatted}\n")

    # Remove stale .apl files
    for existing in os.listdir(ws_dir):
        if existing.endswith(".apl") and existing not in written_files:
            os.unlink(os.path.join(ws_dir, existing))


def load_workspace(env: dict[str, Any], ws_dir: str) -> None:
    """Load workspace from a directory."""
    # Read .ws marker for WSID
    ws_file = os.path.join(ws_dir, ".ws")
    if os.path.isfile(ws_file):
        with open(ws_file) as f:
            wsid = f.readline().strip()
            env["__wsid__"] = wsid
            env["⎕WSID"] = APLArray([len(wsid)], list(wsid))

    # Collect .apl files, system vars first
    files = sorted(os.listdir(ws_dir))
    sys_files = [f for f in files if f.startswith("__") and f.endswith(".apl")]
    user_files = [f for f in files if not f.startswith("__") and f.endswith(".apl")]

    for filename in sys_files + user_files:
        filepath = os.path.join(ws_dir, filename)
        with open(filepath) as f:
            line = f.read().strip()
            if line:
                interpret(line, env)


def list_workspaces(root: str) -> list[str]:
    """List workspace names under the given root directory."""
    if not os.path.isdir(root):
        return []
    result = []
    for name in sorted(os.listdir(root)):
        ws_dir = os.path.join(root, name)
        if os.path.isdir(ws_dir) and os.path.isfile(os.path.join(ws_dir, ".ws")):
            result.append(name)
    return result




# Backtick prefix map: `x → APL glyph
# Layout follows Dyalog's backtick keyboard where possible
GLYPH_MAP: dict[str, str] = {
    # Arithmetic
    "-": "×",
    "=": "÷",
    # Comparison
    "<": "≤",
    ">": "≥",
    "/": "≠",
    # Structural
    "r": "⍴",
    "i": "⍳",
    "e": "∈",
    "t": "↑",
    "y": "↓",
    "q": "⌽",
    "Q": "⍉",
    # Grade
    "g": "⍋",
    "G": "⍒",
    # Logic
    "^": "∧",
    "v": "∨",
    "~": "⍲",
    "T": "⍱",
    # Math
    "*": "⍟",
    "o": "○",
    "!": "⌈",
    "d": "⌊",
    "D": "⌹",
    "p": "⌈",
    "b": "⌊",
    # Encode/decode
    "n": "⊤",
    "N": "⊥",
    # I/O and special
    "w": "⍵",
    "a": "⍺",
    "V": "∇",
    "l": "←",
    "x": "⋄",
    "z": "⍎",
    "Z": "⍕",
    # Jot, compose, rank, from
    "j": "∘",
    "J": "⍤",
    "I": "⌷",
    "B": "⌶",
    # High minus
    "2": "¯",
    # Comment
    "c": "⍝",
    # Braces (convenience)
    "[": "{",
    "]": "}",
}


def expand_glyphs(line: str) -> str:
    """Replace backtick sequences with APL glyphs.

    `x → APL glyph, `` → literal backtick.
    """
    result: list[str] = []
    i = 0
    while i < len(line):
        if line[i] == "`" and i + 1 < len(line):
            next_ch = line[i + 1]
            if next_ch == "`":
                result.append("`")
                i += 2
            elif next_ch in GLYPH_MAP:
                result.append(GLYPH_MAP[next_ch])
                i += 2
            else:
                result.append("`")
                i += 1
        else:
            result.append(line[i])
            i += 1
    return "".join(result)



"""Serial bridge to a Pico running MARPLE."""

import time

import serial


class PicoConnection:
    """Wraps a serial connection to a Pico running MARPLE."""

    SENTINEL = "\x00"

    def __init__(self, port: str, baud: int = 115200, timeout: float = 10) -> None:
        self.ser = serial.Serial(port, baud, timeout=timeout)
        time.sleep(1)
        self._drain()

    def _drain(self) -> None:
        """Discard any buffered output from the Pico."""
        while self.ser.in_waiting:
            self.ser.read(self.ser.in_waiting)
            time.sleep(0.1)

    def eval(self, expr: str) -> str:
        """Send an APL expression and return the response text.

        Returns the joined output lines (excluding the sentinel and
        the hex echo). Raises TimeoutError if no sentinel is received.
        """
        encoded = expr.encode("utf-8").hex()
        self.ser.write((encoded + "\r\n").encode("ascii"))
        self.ser.flush()

        lines: list[str] = []
        while True:
            raw = self.ser.readline()
            if not raw:
                raise TimeoutError(
                    f"Pico did not respond to: {expr!r}\n"
                    f"Partial output: {lines}"
                )
            line = raw.decode("utf-8", errors="replace").rstrip("\r\n")
            if line == self.SENTINEL:
                break
            if line == encoded:
                continue
            lines.append(line)
        return "\n".join(lines)

    def eval_silent(self, expr: str) -> None:
        """Send an expression that produces no output (e.g. assignment).

        Waits for the sentinel. Raises AssertionError if the Pico
        returns an error.
        """
        result = self.eval(expr)
        assert "ERROR" not in result, f"Unexpected error for {expr!r}: {result}"

    def close(self) -> None:
        self.ser.close()




try:
    from typing import Any
except ImportError:
    pass

import os
import sys

from marple.arraymodel import APLArray
from marple.errors import APLError
from marple.interpreter import _DfnClosure, default_env, interpret
from marple.parser import Assignment, Program, parse
try:
    from marple.terminal import read_line
except ImportError:
    read_line = None  # type: ignore[assignment]
try:
    from marple.workspace import save_workspace, load_workspace, list_workspaces
except ImportError:
    save_workspace = load_workspace = list_workspaces = None  # type: ignore[assignment]

try:
    WORKSPACES_ROOT = os.environ.get("MARPLE_WORKSPACES", "workspaces")
except AttributeError:
    WORKSPACES_ROOT = "workspaces"


def _rjust(s, width):
    if len(s) >= width:
        return s
    return " " * (width - len(s)) + s


def _is_char_array(arr: APLArray) -> bool:
    return len(arr.data) > 0 and all(isinstance(x, str) for x in arr.data)


from marple.formatting import format_num as _format_num


def format_result(result: APLArray, env: dict[str, Any] | None = None) -> str:
    pp = 10
    if env is not None:
        pp_val = env.get("⎕PP")
        if pp_val is not None:
            pp = int(pp_val.data[0])
    if result.is_scalar():
        return _format_num(result.data[0], pp)
    if _is_char_array(result):
        if len(result.shape) == 1:
            return "".join(str(x) for x in result.data)
        if len(result.shape) == 2:
            rows, cols = result.shape
            lines = []
            for r in range(rows):
                row_data = result.data[r * cols : (r + 1) * cols]
                lines.append("".join(str(x) for x in row_data))
            return "\n".join(lines)
    if len(result.shape) == 1:
        return " ".join(_format_num(x, pp) for x in result.data)
    if len(result.shape) == 2:
        rows, cols = result.shape
        strs = [_format_num(result.data[r * cols + c], pp) for r in range(rows) for c in range(cols)]
        col_widths = []
        for c in range(cols):
            w = max(len(strs[r * cols + c]) for r in range(rows))
            col_widths.append(w)
        lines = []
        for r in range(rows):
            parts = []
            for c in range(cols):
                parts.append(_rjust(strs[r * cols + c], col_widths[c]))
            lines.append(" ".join(parts))
        return "\n".join(lines)
    return repr(result)


def _is_silent(line: str) -> bool:
    """Check if a line is a comment, bare assignment, or directive (should not print)."""
    stripped = line.strip()
    if stripped.startswith("#"):
        return True
    if stripped.startswith("⍝"):
        return True
    try:
        tree = parse(line)
    except Exception:
        return False
    if isinstance(tree, Assignment):
        return True
    if isinstance(tree, Program):
        return len(tree.statements) > 0 and isinstance(tree.statements[-1], Assignment)
    return False


def _user_names(env: dict[str, Any]) -> list[str]:
    return sorted(
        name for name in env
        if not name.startswith("⎕") and not name.startswith("__")
        and name not in ("⍵", "⍺", "∇")
    )


def main() -> None:
    # Check for script mode: marple script.marple
    if len(sys.argv) > 1:
        from marple.script import run_script
        path = sys.argv[1]
        for line in run_script(path):
            print(line)
        return

    env: dict[str, Any] = default_env()
    from marple import __version__ as ver
    print(f"MARPLE v{ver} - Mini APL in Python")
    print("CLEAR WS\n")
    use_terminal = read_line is not None and sys.stdin.isatty()
    while True:
        if use_terminal:
            line = read_line()
        else:
            try:
                line = input("      ")
            except (EOFError, KeyboardInterrupt):
                print()
                break
        if line is None:
            break
        line = line.strip()
        if not line:
            continue
        if line == ")off":
            break
        if line == ")clear":
            wsid = env.get("__wsid__", "CLEAR WS")
            env.clear()
            env["__wsid__"] = "CLEAR WS"
            print("CLEAR WS")
            continue
        if line == ")wsid" or line == ")WSID":
            print(env.get("__wsid__", "CLEAR WS"))
            continue
        if line.startswith(")wsid ") or line.startswith(")WSID "):
            new_wsid = line.split(None, 1)[1].strip()
            env["__wsid__"] = new_wsid
            print(new_wsid)
            continue
        if line.startswith(")fns"):
            parts = line.split(None, 1)
            if len(parts) > 1:
                # )fns $::str or )fns utils
                ns_name = parts[1].strip()
                from marple.interpreter import _get_sys_workspace
                if ns_name.startswith("$"):
                    ns_parts = ns_name.split("::")[1:] if "::" in ns_name else []
                    sys_ws = _get_sys_workspace()
                    ns = sys_ws.resolve(ns_parts) if ns_parts else sys_ws
                    if ns is not None and hasattr(ns, "list_names"):
                        print("  ".join(ns.list_names()))
                    else:
                        print(f"Namespace not found: {ns_name}")
                else:
                    print(f"Namespace not found: {ns_name}")
            else:
                fns = [n for n in _user_names(env) if isinstance(env[n], _DfnClosure)]
                print("  ".join(fns) if fns else "")
            continue
        if line == ")vars":
            vars_ = [n for n in _user_names(env) if isinstance(env[n], APLArray)]
            print("  ".join(vars_) if vars_ else "")
            continue
        if line == ")lib" or line == ")LIB":
            workspaces = list_workspaces(WORKSPACES_ROOT)
            if workspaces:
                print("  ".join(workspaces))
            else:
                print("(none)")
            continue
        if line.startswith(")save") or line.startswith(")SAVE"):
            parts = line.split(None, 1)
            if len(parts) > 1:
                env["__wsid__"] = parts[1].strip()
            wsid = env.get("__wsid__", "CLEAR WS")
            if wsid == "CLEAR WS":
                print("ERROR: No workspace ID set. Use )WSID name first.")
                continue
            ws_dir = os.path.join(WORKSPACES_ROOT, wsid)
            try:
                save_workspace(env, ws_dir)
                print(f"{wsid} SAVED")
            except Exception as e:
                print(f"ERROR: {e}")
            continue
        if line.startswith(")load") or line.startswith(")LOAD"):
            parts = line.split(None, 1)
            if len(parts) < 2:
                print("ERROR: )LOAD requires a workspace name")
                continue
            name = parts[1].strip()
            ws_dir = os.path.join(WORKSPACES_ROOT, name)
            if not os.path.isdir(ws_dir):
                print(f"ERROR: Workspace not found: {name}")
                continue
            env.clear()
            try:
                load_workspace(env, ws_dir)
                print(env.get("__wsid__", name))
            except Exception as e:
                print(f"ERROR: {e}")
            continue
        try:
            result = interpret(line, env)
            if not _is_silent(line):
                print(format_result(result, env))
        except APLError as e:
            print(e)
        except Exception as e:
            print(f"ERROR: {e}")


if __name__ == "__main__":
    main()



from marple.arraymodel import APLArray, S
from marple.interpreter import interpret, default_env


class TestComparisonTolerance:
    def test_default_ct(self) -> None:
        assert interpret("⎕CT") == S(1e-14)

    def test_near_equal(self) -> None:
        # (1÷3)×3 should equal 1 with tolerant comparison
        assert interpret("1=(1÷3)×3") == S(1)

    def test_far_not_equal(self) -> None:
        assert interpret("1=1.1") == S(0)

    def test_ct_zero_exact(self) -> None:
        env = default_env()
        interpret("⎕CT←0", env)
        # With exact comparison, 1=1.001 should be 0
        assert interpret("1=1.001", env) == S(0)

    def test_near_less_equal(self) -> None:
        env = default_env()
        interpret("x←(1÷3)×3", env)
        assert interpret("1≤x", env) == S(1)

    def test_near_greater_equal(self) -> None:
        env = default_env()
        interpret("x←(1÷3)×3", env)
        assert interpret("x≥1", env) == S(1)

    def test_near_not_equal(self) -> None:
        # Nearly equal values should give 0 for ≠
        assert interpret("1≠(1÷3)×3") == S(0)


class TestIndexOfTolerant:
    def test_iota_tolerant(self) -> None:
        env = default_env()
        interpret("x←÷⍳3", env)  # x is 1 0.5 0.3333...
        # 0.5 should be found at position 2
        assert interpret("x⍳0.5", env) == S(2)

    def test_iota_tolerant_third(self) -> None:
        env = default_env()
        interpret("x←÷⍳3", env)
        # ÷3 (0.3333...) should match x[3] which is also ÷3
        assert interpret("x⍳÷3", env) == S(3)


class TestMembership:
    def test_membership_found(self) -> None:
        assert interpret("3∈1 2 3 4 5") == S(1)

    def test_membership_not_found(self) -> None:
        assert interpret("6∈1 2 3 4 5") == S(0)

    def test_membership_vector(self) -> None:
        assert interpret("1 3 5∈2 3 4") == APLArray([3], [0, 1, 0])

    def test_membership_tolerant(self) -> None:
        env = default_env()
        interpret("x←(1÷3)×3", env)  # nearly 1
        assert interpret("x∈1 2 3", env) == S(1)


class TestMatchIgnoresCT:
    def test_match_exact(self) -> None:
        # ≡ should NOT use ⎕CT — exact comparison
        # This test verifies ≡ is stricter than =
        # Two values that are = (tolerant) might not be ≡ (exact)
        assert interpret("1≡1") == S(1)
        assert interpret("1≡2") == S(0)




import math
try:
    from typing import Callable
except ImportError:
    pass

from marple.arraymodel import APLArray
from marple.backend import (
    _DOWNCAST_CT, _OVERFLOW_UFUNCS, is_numeric_array,
    maybe_downcast, maybe_upcast, np, to_bool_array, to_list,
)
from marple.errors import DomainError, LengthError


def _pervade_monadic(
    f: Callable[[int | float], int | float],
    omega: APLArray,
    ufunc_name: str | None = None,
    bool_result: bool = False,
) -> APLArray:
    if ufunc_name and is_numeric_array(omega.data):
        ufunc = getattr(np, ufunc_name, None)
        if ufunc is not None:
            result = ufunc(omega.data)
            if bool_result:
                result = to_bool_array(result)
            return APLArray(list(omega.shape), result)
    data = [f(x) for x in to_list(omega.data)]
    if bool_result:
        data = to_bool_array(data)
    return APLArray(list(omega.shape), data)


def _pervade_dyadic(
    f: Callable[[int | float, int | float], int | float],
    alpha: APLArray,
    omega: APLArray,
    ufunc_name: str | None = None,
    bool_result: bool = False,
) -> APLArray:
    if (
        ufunc_name
        and is_numeric_array(alpha.data)
        and is_numeric_array(omega.data)
    ):
        ufunc = getattr(np, ufunc_name, None)
        if ufunc is not None:
            a_arr = alpha.data
            b_arr = omega.data
            if ufunc_name in _OVERFLOW_UFUNCS:
                a_arr = maybe_upcast(a_arr)
                b_arr = maybe_upcast(b_arr)
            try:
                result = ufunc(a_arr, b_arr)
            except ValueError:
                raise LengthError(f"Shape mismatch: {alpha.shape} vs {omega.shape}")
            if ufunc_name in _OVERFLOW_UFUNCS:
                result = maybe_downcast(result, _DOWNCAST_CT)
            if bool_result:
                result = to_bool_array(result)
            shape = list(omega.shape) if not omega.is_scalar() else list(alpha.shape)
            return APLArray(shape, result)
    # Fallback: element-wise Python
    a_data = to_list(alpha.data)
    b_data = to_list(omega.data)
    if alpha.is_scalar() and omega.is_scalar():
        return APLArray([], [f(a_data[0], b_data[0])])
    if alpha.is_scalar():
        a = a_data[0]
        data = to_bool_array([f(a, x) for x in b_data]) if bool_result else [f(a, x) for x in b_data]
        return APLArray(list(omega.shape), data)
    if omega.is_scalar():
        b = b_data[0]
        data = to_bool_array([f(x, b) for x in a_data]) if bool_result else [f(x, b) for x in a_data]
        return APLArray(list(alpha.shape), data)
    if alpha.shape != omega.shape:
        raise LengthError(f"Shape mismatch: {alpha.shape} vs {omega.shape}")
    data = [f(a, b) for a, b in zip(a_data, b_data)]
    if bool_result:
        data = to_bool_array(data)
    return APLArray(list(alpha.shape), data)


# Monadic functions

def negate(omega: APLArray) -> APLArray:
    return _pervade_monadic(lambda x: -x, omega, "negative")


def reciprocal(omega: APLArray) -> APLArray:
    def _recip(x: int | float) -> int | float:
        if x == 0:
            raise DomainError("Division by zero")
        return 1 / x
    return _pervade_monadic(_recip, omega)


def ceiling(omega: APLArray) -> APLArray:
    return _pervade_monadic(lambda x: math.ceil(x), omega, "ceil")


def floor(omega: APLArray) -> APLArray:
    return _pervade_monadic(lambda x: math.floor(x), omega, "floor")


# Dyadic functions

def add(alpha: APLArray, omega: APLArray) -> APLArray:
    return _pervade_dyadic(lambda a, b: a + b, alpha, omega, "add")


def subtract(alpha: APLArray, omega: APLArray) -> APLArray:
    return _pervade_dyadic(lambda a, b: a - b, alpha, omega, "subtract")


def multiply(alpha: APLArray, omega: APLArray) -> APLArray:
    return _pervade_dyadic(lambda a, b: a * b, alpha, omega, "multiply")


def divide(alpha: APLArray, omega: APLArray) -> APLArray:
    def _div(a: int | float, b: int | float) -> int | float:
        if b == 0:
            raise DomainError("Division by zero")
        return a / b
    return _pervade_dyadic(_div, alpha, omega)


def maximum(alpha: APLArray, omega: APLArray) -> APLArray:
    return _pervade_dyadic(lambda a, b: max(a, b), alpha, omega, "maximum")


def minimum(alpha: APLArray, omega: APLArray) -> APLArray:
    return _pervade_dyadic(lambda a, b: min(a, b), alpha, omega, "minimum")


# Extended monadic functions

def exponential(omega: APLArray) -> APLArray:
    return _pervade_monadic(lambda x: math.exp(x), omega, "exp")


def natural_log(omega: APLArray) -> APLArray:
    return _pervade_monadic(lambda x: math.log(x), omega, "log")


def absolute_value(omega: APLArray) -> APLArray:
    return _pervade_monadic(lambda x: abs(x), omega, "absolute")


def logical_not(omega: APLArray) -> APLArray:
    return _pervade_monadic(lambda x: int(not x), omega, "logical_not", bool_result=True)


def pi_times(omega: APLArray) -> APLArray:
    return _pervade_monadic(lambda x: math.pi * x, omega)


_CIRCULAR: dict[int, Callable[[float], float]] = {
    0: lambda x: math.sqrt(1 - x * x),
    1: math.sin,
    2: math.cos,
    3: math.tan,
    4: lambda x: math.sqrt(1 + x * x),
    5: math.sinh,
    6: math.cosh,
    7: math.tanh,
    -1: math.asin,
    -2: math.acos,
    -3: math.atan,
    -4: lambda x: math.sqrt(x * x - 1),
    -5: math.asinh,
    -6: math.acosh,
    -7: math.atanh,
}


def circular(alpha: APLArray, omega: APLArray) -> APLArray:
    def _apply(a: int | float, b: int | float) -> int | float:
        fn = _CIRCULAR.get(int(a))
        if fn is None:
            raise DomainError(f"Unknown circular function selector: {a}")
        return fn(float(b))
    return _pervade_dyadic(_apply, alpha, omega)


# Extended dyadic functions

def power(alpha: APLArray, omega: APLArray) -> APLArray:
    return _pervade_dyadic(lambda a, b: a ** b, alpha, omega, "power")


def logarithm(alpha: APLArray, omega: APLArray) -> APLArray:
    return _pervade_dyadic(lambda a, b: math.log(b) / math.log(a), alpha, omega)


def residue(alpha: APLArray, omega: APLArray) -> APLArray:
    return _pervade_dyadic(lambda a, b: b % a, alpha, omega)


def _tolerant_eq(a: int | float, b: int | float, ct: float) -> bool:
    """APL tolerant equality: |a-b| ≤ ct × (|a| ⌈ |b|)"""
    if ct == 0:
        return a == b
    return abs(a - b) <= ct * max(abs(a), abs(b))


def less_than(alpha: APLArray, omega: APLArray, ct: float = 0) -> APLArray:
    return _pervade_dyadic(lambda a, b: int(a < b and not _tolerant_eq(a, b, ct)), alpha, omega, bool_result=True)


def less_equal(alpha: APLArray, omega: APLArray, ct: float = 0) -> APLArray:
    return _pervade_dyadic(lambda a, b: int(a <= b or _tolerant_eq(a, b, ct)), alpha, omega, bool_result=True)


def equal(alpha: APLArray, omega: APLArray, ct: float = 0) -> APLArray:
    return _pervade_dyadic(lambda a, b: int(_tolerant_eq(a, b, ct)), alpha, omega, bool_result=True)


def greater_equal(alpha: APLArray, omega: APLArray, ct: float = 0) -> APLArray:
    return _pervade_dyadic(lambda a, b: int(a >= b or _tolerant_eq(a, b, ct)), alpha, omega, bool_result=True)


def greater_than(alpha: APLArray, omega: APLArray, ct: float = 0) -> APLArray:
    return _pervade_dyadic(lambda a, b: int(a > b and not _tolerant_eq(a, b, ct)), alpha, omega, bool_result=True)


def not_equal(alpha: APLArray, omega: APLArray, ct: float = 0) -> APLArray:
    return _pervade_dyadic(lambda a, b: int(not _tolerant_eq(a, b, ct)), alpha, omega, bool_result=True)


def logical_and(alpha: APLArray, omega: APLArray) -> APLArray:
    return _pervade_dyadic(lambda a, b: int(bool(a) and bool(b)), alpha, omega, "logical_and", bool_result=True)


def logical_or(alpha: APLArray, omega: APLArray) -> APLArray:
    return _pervade_dyadic(lambda a, b: int(bool(a) or bool(b)), alpha, omega, "logical_or", bool_result=True)



from marple.arraymodel import APLArray, S
from marple.errors import LengthError
from marple.interpreter import interpret, default_env
import pytest


class TestRoll:
    def test_roll_deterministic(self) -> None:
        env = default_env()
        interpret("⎕RL←42", env)
        r1 = interpret("?6", env)
        interpret("⎕RL←42", env)
        r2 = interpret("?6", env)
        assert r1 == r2

    def test_roll_in_range(self) -> None:
        env = default_env()
        interpret("⎕RL←42", env)
        for _ in range(20):
            result = interpret("?6", env)
            v = result.data[0]
            assert 1 <= v <= 6

    def test_roll_zero_gives_float(self) -> None:
        env = default_env()
        interpret("⎕RL←42", env)
        result = interpret("?0", env)
        v = result.data[0]
        assert isinstance(v, float)
        assert 0 <= v < 1

    def test_roll_pervades(self) -> None:
        env = default_env()
        interpret("⎕RL←42", env)
        result = interpret("?6 6 6", env)
        assert result.shape == [3]
        for v in result.data:
            assert 1 <= v <= 6

    def test_roll_respects_io(self) -> None:
        env = default_env()
        interpret("⎕IO←0", env)
        interpret("⎕RL←42", env)
        for _ in range(20):
            result = interpret("?6", env)
            v = result.data[0]
            assert 0 <= v <= 5


class TestDeal:
    def test_deal_length(self) -> None:
        env = default_env()
        interpret("⎕RL←42", env)
        result = interpret("5?10", env)
        assert result.shape == [5]

    def test_deal_distinct(self) -> None:
        env = default_env()
        interpret("⎕RL←42", env)
        result = interpret("5?10", env)
        values = list(result.data)
        assert len(set(values)) == 5

    def test_deal_in_range(self) -> None:
        env = default_env()
        interpret("⎕RL←42", env)
        result = interpret("5?10", env)
        for v in result.data:
            assert 1 <= v <= 10

    def test_deal_deterministic(self) -> None:
        env = default_env()
        interpret("⎕RL←42", env)
        r1 = interpret("5?52", env)
        interpret("⎕RL←42", env)
        r2 = interpret("5?52", env)
        assert r1 == r2

    def test_deal_respects_io(self) -> None:
        env = default_env()
        interpret("⎕IO←0", env)
        interpret("⎕RL←42", env)
        result = interpret("5?10", env)
        for v in result.data:
            assert 0 <= v <= 9

    def test_deal_n_exceeds_m_error(self) -> None:
        env = default_env()
        with pytest.raises(LengthError):
            interpret("10?5", env)


class TestQuadRL:
    def test_default_rl(self) -> None:
        assert interpret("⎕RL") == S(1)

    def test_set_rl(self) -> None:
        env = default_env()
        interpret("⎕RL←123", env)
        assert interpret("⎕RL", env) == S(123)



from marple.parser import (
    Assignment,
    DyadicFunc,
    MonadicFunc,
    Num,
    Vector,
    parse,
)


class TestParserNumbers:
    def test_single_integer(self) -> None:
        assert parse("5") == Num(5)

    def test_negative_number(self) -> None:
        assert parse("¯3") == Num(-3)

    def test_float(self) -> None:
        assert parse("3.14") == Num(3.14)


class TestParserVectors:
    def test_strand_notation(self) -> None:
        assert parse("1 2 3") == Vector([Num(1), Num(2), Num(3)])


class TestParserMonadicFunctions:
    def test_negate(self) -> None:
        assert parse("-5") == MonadicFunc("-", Num(5))

    def test_negate_vector(self) -> None:
        assert parse("-1 2 3") == MonadicFunc("-", Vector([Num(1), Num(2), Num(3)]))


class TestParserDyadicFunctions:
    def test_addition(self) -> None:
        assert parse("3+4") == DyadicFunc("+", Num(3), Num(4))

    def test_right_to_left(self) -> None:
        # 1+2×3 should parse as 1+(2×3)
        assert parse("1+2×3") == DyadicFunc("+", Num(1), DyadicFunc("×", Num(2), Num(3)))

    def test_subtraction(self) -> None:
        assert parse("5-3") == DyadicFunc("-", Num(5), Num(3))


class TestParserParentheses:
    def test_parens_override_scope(self) -> None:
        # (1+2)×3 should parse as (1+2)×3
        assert parse("(1+2)×3") == DyadicFunc(
            "×", DyadicFunc("+", Num(1), Num(2)), Num(3)
        )


class TestParserAssignment:
    def test_simple_assignment(self) -> None:
        assert parse("x←5") == Assignment("x", Num(5))

    def test_assignment_with_expression(self) -> None:
        assert parse("x←3+4") == Assignment("x", DyadicFunc("+", Num(3), Num(4)))
